import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import read_csv


class MSDWorkbenchReader:
    SUPPORTED_EXTENSIONS = "csv, txt, xlsx"
    plate_data: pd.DataFrame
    well_plate_id: str

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        if named_file_contents.extension == "xlsx":
            # Read the Workbench data sheet from Excel file
            data = pd.read_excel(
                named_file_contents.contents,
                sheet_name="Workbench data",
                engine="openpyxl",
            )
            data = data.where(pd.notna(data), None)
            data.index = pd.Index(data.index.to_series().ffill())
            data.index = data.index.astype(str).str.strip()
            data.columns = data.columns.astype(str).str.strip()
            first_row = str(data.iloc[0].index[0])
            if "Plate" not in first_row:
                msg = "Plate ID not found in the first row of the data"
                raise AllotropeConversionError(msg)
            self.well_plate_id = first_row.split("_")[-1].strip()
            # Set the first row as the header
            data.columns = pd.Index(data.iloc[0])
            data = data[1:].reset_index(drop=True)
        else:
            # First, peek at the first line to check format
            named_file_contents.contents.seek(0)
            first_line = named_file_contents.contents.readline()
            if isinstance(first_line, bytes):
                first_line = first_line.decode("utf-8")
            named_file_contents.contents.seek(0)

            # Check if first line has commas (old format) or not (new format)
            if "," in first_line:
                # Old format: Plate_ID,,,,,,,,
                data = read_csv(named_file_contents.contents)
                data = data.where(pd.notna(data), None)
                data.index = pd.Index(data.index.to_series().ffill())
                data.index = data.index.astype(str).str.strip()
                data.columns = data.columns.astype(str).str.strip()
                first_row = str(data.iloc[0].index[0])
                if "Plate" not in first_row:
                    msg = "Plate ID not found in the first row of the data"
                    raise AllotropeConversionError(msg)
                self.well_plate_id = first_row.split("_")[-1].strip()
                # Set the first row as the header
                data.columns = pd.Index(data.iloc[0])
                data = data[1:].reset_index(drop=True)
            else:
                # New format: Plate_ID (no commas on first line)
                # Extract plate ID from first line
                plate_id_line = first_line.strip()
                if "Plate" not in plate_id_line:
                    msg = "Plate ID not found in the first row of the data"
                    raise AllotropeConversionError(msg)
                self.well_plate_id = plate_id_line.split("_")[-1].strip()

                # Now read the CSV starting from the second line (which has headers)
                named_file_contents.contents.seek(0)
                # Skip the first line
                named_file_contents.contents.readline()
                # Read rest of file with second line as header
                data = read_csv(named_file_contents.contents)
                data = data.where(pd.notna(data), None)
                data.columns = data.columns.astype(str).str.strip()

        self.plate_data = data



from functools import partial

from allotropy.allotrope.models.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.thermo_fisher_nanodrop_eight.nanodrop_eight_reader import (
    NanodropEightReader,
)
from allotropy.parsers.thermo_fisher_nanodrop_eight.nanodrop_eight_structure import (
    create_measurement_group,
    create_metadata,
    SpectroscopyRow,
)
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.vendor_parser import VendorParser


class NanodropEightParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Thermo Fisher Scientific NanoDrop Eight"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = NanodropEightReader.SUPPORTED_EXTENSIONS

    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = NanodropEightReader(named_file_contents)
        rows = map_rows(
            reader.data, partial(SpectroscopyRow.create, header=reader.header)
        )
        metadata = create_metadata(
            reader.header, named_file_contents.original_file_path
        )

        return Data(
            metadata=metadata,
            measurement_groups=[
                create_measurement_group(
                    row, reader.header, SeriesData(reader.data.iloc[0])
                )
                for row in rows
            ],
            # NOTE: in current implementation, calculated data is reported at global level for some reason.
            # TODO(nstender): should we move this inside of measurements?
            calculated_data=[item for row in rows for item in row.calculated_data],
        )



import numpy as np
import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import read_csv


class QiacuitydPCRReader:
    SUPPORTED_EXTENSIONS = "csv"
    well_data: pd.DataFrame

    def __init__(self, named_file_contents: NamedFileContents):
        # Read in the file, skip first row since it does not have data in it.
        qiacuity_dpcr_data = read_csv(
            filepath_or_buffer=named_file_contents.contents,
            header=1,
            encoding=named_file_contents.encoding,
        ).replace(np.nan, None)
        qiacuity_dpcr_data.columns = qiacuity_dpcr_data.columns.str.replace("�", "μ")
        qiacuity_dpcr_data.columns = qiacuity_dpcr_data.columns.str.replace("Âμ", "μ")
        column_names = qiacuity_dpcr_data.columns.tolist()
        if "Well" in column_names:
            # Rename "Well" to "Well Name" for consistency
            column_names = [
                "Well Name" if col == "Well" else col for col in column_names
            ]
        else:
            # Rename the blank first column to specify that it's the Well Name column
            column_names[0] = "Well Name"
        column_index = pd.Index(column_names)
        qiacuity_dpcr_data.columns = column_index
        self.well_data = qiacuity_dpcr_data



from io import BytesIO

import numpy as np
import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import (
    CsvReader,
    EMPTY_STR_OR_CSV_LINE,
)
from allotropy.parsers.utils.values import assert_not_none
from allotropy.parsers.utils.zip_handler import ZipHandler

# we don't do chained assignment, disable to prevent spurious warning.
pd.options.mode.chained_assignment = None


def get_first_non_empty(data: list[str]) -> int | None:
    for idx, col in enumerate(data):
        if col:
            return idx
    return None


def first_row_as_columns(df: pd.DataFrame) -> pd.DataFrame:
    df.columns = pd.Index(df.iloc[0].astype(str).tolist())
    return df[1:]


class AppbioAbsoluteQReader:
    SUPPORTED_EXTENSIONS = "csv,zip"
    data: pd.DataFrame
    common_columns: list[str]

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        file_contents = (
            NamedFileContents(
                contents=self._parse_zip_contents(
                    named_file_contents.get_bytes_stream()
                ),
                original_file_path=named_file_contents.original_file_path,
                encoding=named_file_contents.encoding,
            )
            if named_file_contents.extension == "zip"
            else named_file_contents
        )

        csv_reader = CsvReader.create(file_contents)
        df, self.common_columns = self.parse_dataframe(csv_reader)

        columns_to_rename = {}
        if "Name" in df and "Sample" not in df:
            columns_to_rename["Name"] = "Sample"
        if "Well ID" in df and "Well" not in df:
            columns_to_rename["Well ID"] = "Well"
        if "Run name" in df and "Run" not in df:
            columns_to_rename["Run name"] = "Run"

        if columns_to_rename:
            df = df.rename(columns=columns_to_rename)

        required_keys = {"Sample"}
        for key in required_keys:
            if key not in df:
                possible_keys = key
                if key in columns_to_rename:
                    possible_keys += f" or {columns_to_rename[key]}"
                msg = f"Input is missing required column '{possible_keys}'"
                raise AllotropeConversionError(msg)

        self.data = df.replace(np.nan, None)

    def _parse_zip_contents(self, data: BytesIO) -> BytesIO:
        zip_handler = ZipHandler(data)

        summary_files = [
            name for name in zip_handler.name_list if name.endswith("_summary.csv")
        ]
        n_files = len(summary_files)

        if n_files > 1:
            msg = f"Multiple summary files found in zip folder - the adapter requires exactly one summary file to be present. File list: {zip_handler.name_list}"
            raise AllotropeConversionError(msg)
        elif n_files == 0:
            msg = f"No summary file identified in zip folder. File list: {zip_handler.name_list}"
            raise AllotropeConversionError(msg)

        return zip_handler.get_file(summary_files[0])

    def parse_dataframe(self, csv_reader: CsvReader) -> tuple[pd.DataFrame, list[str]]:
        titles = assert_not_none(
            csv_reader.pop(),
            msg="Can not parse empty file",
        ).split(",")

        info = assert_not_none(
            csv_reader.get(),
            msg="Can not parse empty file",
        ).split(",")

        first_non_empty = assert_not_none(
            get_first_non_empty(titles),
            msg="Can not parse empty file",
        )

        # A summary file only contains dye settings in the first row
        # The second row contains columns with dye settings info starting from Target until Threshold
        # If not a summary file, parse the csv as is
        if info[first_non_empty] != "Target":
            return (
                assert_not_none(
                    csv_reader.pop_csv_block_as_df(EMPTY_STR_OR_CSV_LINE, names=titles),
                    "Failed to parser dataframe from file.",
                ),
                [],
            )

        df = assert_not_none(
            csv_reader.pop_csv_block_as_df(EMPTY_STR_OR_CSV_LINE, header=None),
            "Failed to parser dataframe from file.",
        )

        dye_names = [column for column in titles if column]

        # Get the start and end indices of the dye setting sections.
        dye_start = []
        dye_end = []
        for idx, col in enumerate(info):
            if col == "Target":
                dye_start.append(idx)
            elif col == "Threshold":
                dye_end.append(idx)

        if len(dye_start) != len(dye_end) != len(dye_names):
            msg = "Unable to infer dye columns"
            raise AllotropeConversionError(msg)

        # expected size of dye settings
        dye_size = dye_end[0] - dye_start[0] + 1

        # Construct common columns
        start, end = dye_start[0], dye_end[-1]
        base_df = first_row_as_columns(
            pd.concat([df.iloc[:, :start], df.iloc[:, end + 1 :]], axis=1)
        )

        # The group column is left blank on a ffill basis.
        base_df.loc[:, "Group"] = base_df["Group"].ffill()

        # Extract each dye setting columns and combine with the common columns
        # to create a dataset that looks like non-summary files, with one dye setting per row.
        dye_dfs: list[pd.DataFrame] = []
        for start, end, name in zip(dye_start, dye_end, dye_names, strict=True):
            dye_df = first_row_as_columns(df.iloc[:, start : end + 1])

            if dye_df.shape[1] != dye_size:
                msg = f"Expected all length of dye sections to be {dye_size}, dye {name} got: {dye_df.shape[1]}"
                raise AllotropeConversionError(msg)

            dye_df.loc[:, "Dye"] = name
            dye_dfs.append(pd.concat([base_df, dye_df], axis="columns"))

        return pd.concat(dye_dfs, axis="index"), list(base_df.columns)



from __future__ import annotations

from decimal import Decimal
from pathlib import Path
import re
from typing import Any

import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.models.shared.definitions.units import (
    MilliAbsorbanceUnit,
    Nanometer,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2025._03.plate_reader import (
    ErrorDocument,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ProcessedDataDocument,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.constants import NEGATIVE_ZERO, NOT_APPLICABLE
from allotropy.parsers.unchained_labs_lunatic_stunner.constants import (
    CALCULATED_DATA_LOOKUP,
    DEFAULT_DETECTION_TYPE,
    DEVICE_TYPE,
    DYNAMIC_LIGHT_SCATTERING_DETECTION_TYPE,
    INCORRECT_WAVELENGTH_COLUMN_FORMAT_ERROR_MSG,
    MODEL_NUMBER,
    NO_DATE_OR_TIME_ERROR_MSG,
    NO_DEVICE_IDENTIFIER_ERROR_MSG,
    NO_MEASUREMENT_IN_PLATE_ERROR_MSG,
    NO_WAVELENGTH_COLUMN_ERROR_MSG,
    PRODUCT_MANUFACTURER,
    SOFTWARE_NAME,
    WAVELENGTH_COLUMNS_RE,
)
from allotropy.parsers.unchained_labs_lunatic_stunner.unchained_labs_lunatic_stunner_calcdocs import (
    create_calculated_data,
)
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
)
from allotropy.parsers.utils.pandas import (
    map_rows,
    SeriesData,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
)


def _extract_peak_data(well_plate_data: SeriesData) -> list[dict[str, Any]]:
    """Extract peak data from well plate data and return as list of peak info dictionaries."""
    peak_data = []

    peak_pattern = re.compile(r"peak (\d+) ")
    peak_numbers = set()

    for column in well_plate_data.series.index:
        match = peak_pattern.match(str(column).lower())
        if match:
            peak_numbers.add(int(match.group(1)))

    for peak_num in sorted(peak_numbers):
        peak_info: dict[str, Any] = {}

        mean_dia = well_plate_data.get(float, f"peak {peak_num} mean dia (nm)")
        if mean_dia is not None:
            peak_info["peak mean diameter"] = mean_dia

        mode_dia = well_plate_data.get(float, f"peak {peak_num} mode dia (nm)")
        if mode_dia is not None:
            peak_info["peak mode diameter"] = mode_dia

        est_mw = well_plate_data.get(float, f"peak {peak_num} est. mw (kda)")
        if est_mw is not None:
            peak_info["peak est. MW"] = est_mw

        intensity = well_plate_data.get(float, f"peak {peak_num} intensity (%)")
        if intensity is not None:
            peak_info["peak intensity"] = intensity

        mass = well_plate_data.get(float, f"peak {peak_num} mass (%)")
        if mass is not None:
            peak_info["peak mass"] = mass

        peak_info["peak index"] = str(peak_num)

        if len(peak_info) > 1:
            peak_data.append(peak_info)

    return peak_data


def _get_calculated_value_and_is_na(
    series: SeriesData, key: str
) -> tuple[float | None, bool]:
    """Return value and whether the source cell is literal N/A.

    - If the raw cell value is the string literal "N/A", returns (NEGATIVE_ZERO, True).
    - If the cell is empty/NaN, returns (None, False) so caller can skip the key.
    - Otherwise, returns (float value, False).
    """
    raw_string = series.get(str, key, validate=SeriesData.NOT_NAN)
    if raw_string is None:
        return None, False
    stripped = str(raw_string).strip()
    if stripped == "":
        return None, False
    if stripped.upper() == NOT_APPLICABLE:
        return NEGATIVE_ZERO, True
    return series.get(float, key), False


def _is_literal_not_applicable(series: SeriesData, key: str) -> bool:
    """Return True if the source cell is the literal string N/A (case-insensitive)."""
    raw_string = series.get(str, key, validate=SeriesData.NOT_NAN)
    if raw_string is None:
        return False
    return str(raw_string).strip().upper() == NOT_APPLICABLE


def _create_measurement(
    well_plate_data: SeriesData,
    header: SeriesData,
    wavelength_columns: list[str],
) -> Measurement:
    absorbance_errors: list[ErrorDocument] = []
    for wavelength_column in wavelength_columns:
        if wavelength_column not in well_plate_data.series:
            msg = NO_MEASUREMENT_IN_PLATE_ERROR_MSG.format(wavelength_column)
            raise AllotropeConversionError(msg)

        wavelength_match = WAVELENGTH_COLUMNS_RE.match(wavelength_column)
        if not wavelength_match:
            raise AllotropeConversionError(INCORRECT_WAVELENGTH_COLUMN_FORMAT_ERROR_MSG)

        absorbance_is_na = _is_literal_not_applicable(
            well_plate_data, wavelength_column
        )
        if absorbance_is_na:
            absorbance_errors.append(
                ErrorDocument(
                    error=NOT_APPLICABLE,
                    error_feature=f"{DEFAULT_DETECTION_TYPE.lower()}",
                )
            )

    background_wavelength = well_plate_data.get(float, "background wvl. (nm)")
    background_absorbance = None
    if background_wavelength is not None:
        background_absorbance = well_plate_data.get(
            float, f"background (a{int(background_wavelength)})"
        )

    measurement_identifier = random_uuid_str()
    peak_data = _extract_peak_data(well_plate_data)

    error_documents: list[ErrorDocument] = []

    concentration_factor = well_plate_data.get(float, "concentration factor (ng/ul)")
    application = header.get(str, "application")
    analytical_method_identifier = (
        well_plate_data.get(str, "application") or application
    )
    experimental_data_identifier = header.get(str, "experiment name")
    detection_type = DEFAULT_DETECTION_TYPE
    if application and "B22 & kD" in application:
        detection_type = DYNAMIC_LIGHT_SCATTERING_DETECTION_TYPE

    calculated_data_values, calculated_data_errors = _get_calculated_data_values(
        well_plate_data, wavelength_columns
    )

    error_documents.extend(calculated_data_errors)
    error_documents.extend(absorbance_errors)

    # Initialize optional variables to None; set if applicable
    spectrum_data_cube: DataCube | None = None
    absorbance_value: float | None = None
    detector_wavelength_setting: float | None = None
    wavelength_identifier: str | None = None

    if len(wavelength_columns) > 1:
        spectrum_data_cube = _get_spectrum_data_cube(
            well_plate_data, wavelength_columns
        )
    elif len(wavelength_columns) == 1:
        wavelength_match = WAVELENGTH_COLUMNS_RE.match(wavelength_columns[0])
        wavelength, _ = wavelength_match.groups() if wavelength_match else ("", "")
        absorbance_tmp = well_plate_data.get(float, wavelength_columns[0])
        absorbance_value = (
            absorbance_tmp if absorbance_tmp is not None else NEGATIVE_ZERO
        )
        detector_wavelength_setting = float(wavelength) if wavelength else None
        wavelength_identifier = wavelength_columns[0]

    return Measurement(
        type_=MeasurementType.ULTRAVIOLET_ABSORBANCE_CUBE_SPECTRUM
        if len(wavelength_columns) > 1
        else MeasurementType.ULTRAVIOLET_ABSORBANCE,
        device_type=DEVICE_TYPE,
        detection_type=detection_type,
        identifier=measurement_identifier,
        analytical_method_identifier=analytical_method_identifier,
        experimental_data_identifier=experimental_data_identifier,
        detector_wavelength_setting=detector_wavelength_setting,
        electronic_absorbance_reference_wavelength_setting=background_wavelength,
        absorbance=absorbance_value,
        sample_identifier=well_plate_data[str, "sample name"],
        location_identifier=well_plate_data[str, "plate position"],
        well_plate_identifier=well_plate_data.get(str, "plate id"),
        batch_identifier=well_plate_data.get(str, "sample group"),
        firmware_version=header.get(str, "client version"),
        number_of_averages=well_plate_data.get(float, "number of acquisitions"),
        integration_time=well_plate_data.get(float, "acquisition time (s)"),
        compartment_temperature=well_plate_data.get(float, "temperature (°c)"),
        sample_custom_info={
            "plate type": header.get(str, "plate type")
            or well_plate_data.get(str, "plate type"),
            "nr of plates": header.get(str, "nr of plates"),
            "blanks": header.get(str, "blanks"),
            "plate description": header.get(str, "nan", duplicate_strategy="last"),
            "molar attenuation coefficient setting": well_plate_data.get(float, "e1%"),
            "analyte": well_plate_data.get(str, "analyte"),
            "buffer": well_plate_data.get(str, "buffer"),
            "molecular weight (kda)": well_plate_data.get(
                float, "molecular weight (kda)"
            ),
        },
        device_control_custom_info={
            "path length mode": well_plate_data.get(str, "path length mode"),
            "pump": well_plate_data.get(str, "pump"),
            "column": header.get(str, "column") or well_plate_data.get(str, "column"),
            "Number of acquisitions used": well_plate_data.get(
                str, "number of acquisitions used"
            ),
            "Acquisition filtering": well_plate_data.get(str, "acquisition filtering"),
        },
        error_document=error_documents,
        processed_data_document=ProcessedDataDocument(
            identifier=random_uuid_str(),
            concentration_factor=concentration_factor,
            peak_list_custom_info=peak_data,
        )
        if concentration_factor is not None or peak_data
        else None,
        calc_docs_custom_info={
            **calculated_data_values,
            **{
                "b22 linear fit": well_plate_data.get(str, "b22 linear fit"),
                "kd linear fit": well_plate_data.get(str, "kd linear fit"),
            },
        },
        measurement_custom_info={
            "electronic_absorbance_reference_absorbance": background_absorbance,
            **_filter_empty_string_values(
                well_plate_data.get_unread(
                    # Skip already mapped columns from well plate data (repeated in CSV no header cases)
                    skip={
                        "time",
                        "row",
                        "concentration factor (ng/ul)",
                        "application",
                        "concentration (ng/ul)",
                        "background wvl. (nm)",
                        "plate id",
                        "plate position",
                        "plate type",
                        "instrument id",
                        "column",
                        # Strings to skip since these are already captured as measurements/calculated data
                        # Skip absorbance measurements with a### (10mm) -- spectral scans
                        r"^a\d{3} \(10mm\)$",
                        # a###/a### -- calculated purity values
                        r"^a\d{3}/a\d{3}$",
                        # a### -- raw absorbance measurement
                        r"^a\d{3}$",
                    },
                )
            ),
        },
        spectrum_data_cube=spectrum_data_cube,
        wavelength_identifier=wavelength_identifier,
    )


def _get_spectrum_data_cube(
    well_plate_data: SeriesData, wavelength_columns: list[str]
) -> DataCube:
    wavelengths, absorbances = _get_wavelengths_and_absorbance(
        well_plate_data, wavelength_columns
    )

    return DataCube(
        label="absorbance-spectrum",
        structure_dimensions=[
            DataCubeComponent(
                concept="wavelength",
                type_=FieldComponentDatatype.double,
                unit=Nanometer.unit,
            )
        ],
        structure_measures=[
            DataCubeComponent(
                concept="absorbance",
                type_=FieldComponentDatatype.double,
                unit=MilliAbsorbanceUnit.unit,
            )
        ],
        dimensions=[wavelengths],
        measures=[absorbances],
    )


def _get_wavelengths_and_absorbance(
    well_plate_data: SeriesData, wavelength_columns: list[str]
) -> tuple[list[float], list[float]]:
    wavelength_to_absorbance: dict[float, Decimal | None] = {}
    for wavelength_column in wavelength_columns:
        match = WAVELENGTH_COLUMNS_RE.match(wavelength_column)
        if not match:
            raise AllotropeConversionError(INCORRECT_WAVELENGTH_COLUMN_FORMAT_ERROR_MSG)
        wavelength_str, _ = match.groups()
        wavelength = float(wavelength_str)
        # Read as string to preserve precision; fallback to float; allow None
        raw_str = well_plate_data.get(
            str, wavelength_column, validate=SeriesData.NOT_NAN
        )
        absorbance: Decimal | None
        if raw_str is None or str(raw_str).strip() == "":
            absorbance = None
        else:
            absorbance = Decimal(str(raw_str).strip())

        if wavelength in wavelength_to_absorbance:
            previous_absorbance = wavelength_to_absorbance[wavelength]
            if absorbance is None or previous_absorbance is None:
                continue
            else:
                wavelength_to_absorbance[
                    wavelength
                ] = _get_absorbance_with_highest_precision(
                    previous_absorbance, absorbance
                )

        else:
            wavelength_to_absorbance[wavelength] = absorbance

    sorted_wavelengths = sorted(wavelength_to_absorbance.keys())

    absorbances = []
    for w in sorted_wavelengths:
        absorbance = wavelength_to_absorbance[w]
        if absorbance is None:
            absorbances.append(float(NEGATIVE_ZERO))
        else:
            absorbances.append(float(absorbance))

    return sorted_wavelengths, absorbances


def _get_absorbance_with_highest_precision(
    absorbance1: Decimal, absorbance2: Decimal
) -> Decimal:
    def _int_exponent(d: Decimal) -> int:
        exp = d.as_tuple().exponent
        if not isinstance(exp, int):
            msg = "Invalid absorbance value (NaN/Inf) not allowed for precision comparison."
            raise AllotropeConversionError(msg)
        return exp

    exp1 = _int_exponent(absorbance1)
    exp2 = _int_exponent(absorbance2)

    # Determine which has higher precision (more decimal places => smaller exponent),
    # for example, 1.56 has an exponent of -2, while 1.562 has an exponent of -3.
    if exp1 < exp2:
        high_precision = absorbance1
        low_precision, low_precision_exp = absorbance2, exp2
    else:
        high_precision = absorbance2
        low_precision, low_precision_exp = absorbance1, exp1

    # Round both to the lower precision and ensure they match
    quant = Decimal(1).scaleb(low_precision_exp)  # e.g., 1E-2 for two decimal places
    if high_precision.quantize(quant) != low_precision.quantize(quant):
        msg = (
            f"Conflicting absorbance values at same wavelength: {absorbance1} vs {absorbance2} "
            f"when rounded to {abs(low_precision_exp)} decimal places."
        )
        raise AllotropeConversionError(msg)

    # Return the value with the highest precision
    return high_precision


def _get_calculated_data_values(
    well_plate_data: SeriesData, wavelength_columns: list[str]
) -> tuple[dict[str, float], list[ErrorDocument]]:
    calculated_data_values: dict[str, float] = {}
    error_documents: list[ErrorDocument] = []
    for wavelength_column in wavelength_columns:
        for item in CALCULATED_DATA_LOOKUP.get(wavelength_column, []):
            value, is_na = _get_calculated_value_and_is_na(
                well_plate_data, item["column"]
            )
            # Only create error docs when the cell is the literal "N/A"
            if is_na:
                error_documents.append(
                    ErrorDocument(
                        error=NOT_APPLICABLE,
                        error_feature=item["name"],
                    )
                )
            # Skip missing cells entirely; include numeric values
            if value is not None:
                calculated_data_values[item["column"]] = value
    return calculated_data_values, error_documents


def _create_measurement_group(
    data: SeriesData,
    wavelength_columns: list[str],
    header: SeriesData,
) -> MeasurementGroup:
    timestamp = header.get(str, "date")
    # Support timestamp from metadata section, but overide with columns in data if specified.
    date = data.get(str, "date")
    time = data.get(str, "time")
    if date and time:
        timestamp = f"{date} {time}"

    return MeasurementGroup(
        measurement_time=assert_not_none(timestamp, msg=NO_DATE_OR_TIME_ERROR_MSG),
        analyst=header.get(str, "test performed by"),
        plate_well_count=96,
        measurements=[_create_measurement(data, header, wavelength_columns)],
    )


def create_metadata(header: SeriesData, file_path: str) -> Metadata:
    asm_file_identifier = Path(file_path).with_suffix(".json")
    device_identifier = header.get(str, "instrument id") or header.get(
        str, "instrument"
    )
    return Metadata(
        model_number=MODEL_NUMBER,
        product_manufacturer=PRODUCT_MANUFACTURER,
        device_identifier=assert_not_none(
            device_identifier, msg=NO_DEVICE_IDENTIFIER_ERROR_MSG
        ),
        software_name=SOFTWARE_NAME,
        software_version=header.get(str, "software version"),
        file_name=Path(file_path).name,
        unc_path=file_path,
        asm_file_identifier=asm_file_identifier.name,
        data_system_instance_id=NOT_APPLICABLE,
        metadata_custom_info=_filter_empty_string_values(
            header.get_unread(
                # Skip already mapped columns from well plate data (repeated in CSV no header cases)
                skip={
                    "time",
                    "row",
                    "path length mode",
                    "sample group",
                    "pump",
                    "concentration factor (ng/ul)",
                    "sample name",
                    "application",
                    "concentration (ng/ul)",
                    "e1%",
                    "concentration (mg/ml)",
                    "background wvl. (nm)",
                    "plate id",
                    "plate position",
                    # Strings to skip since these are already captured as measurements/calculated data
                    # Skip absorbance spectrum measurements with a### (10mm)
                    r"^a\d{3} \(10mm\)$",
                    # a###/a###
                    r"^a\d{3}/a\d{3}$",
                    # a###
                    r"^a\d{3}$",
                    # background a###
                    r"^background \(a\d{3}\)$",
                    # a### concentration (ng/uL)
                    r"^a\d{3} concentration \(ng/ul\)$",
                },
            )
        ),
    )


def _filter_empty_string_values(unread: dict[str, Any]) -> dict[str, Any]:
    return {
        key: value
        for key, value in unread.items()
        if not (isinstance(value, str) and value.strip() == "")
    }


def create_measurement_groups(
    header: SeriesData, data: pd.DataFrame
) -> tuple[list[MeasurementGroup], list[CalculatedDocument]]:
    wavelength_columns = list(filter(WAVELENGTH_COLUMNS_RE.match, data.columns))
    if not wavelength_columns:
        raise AllotropeConversionError(NO_WAVELENGTH_COLUMN_ERROR_MSG)

    def make_group(data: SeriesData) -> MeasurementGroup:
        return _create_measurement_group(data, wavelength_columns, header)

    measurement_groups = map_rows(data, make_group)
    return measurement_groups, create_calculated_data(measurement_groups)



from __future__ import annotations

from collections.abc import Callable
import functools
import gc
import os
from typing import Any


def suppress_unused_keys_warning(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Use when a class that warns for unread keys is created for a special one-off use that does not intend to read all keys, e.g.
    to read a single value and return.
    Not appropriate to use in the main use of the class where unread keys should be copied into custom data fields.
    """

    @functools.wraps(func)
    def _wrapper(*args: Any, **kwargs: Any) -> Any:
        previous_warn_unused_keys = os.environ.pop("WARN_UNUSED_KEYS", None)
        try:
            return func(*args, **kwargs)
        finally:
            gc.collect()
            if previous_warn_unused_keys is not None:
                os.environ["WARN_UNUSED_KEYS"] = previous_warn_unused_keys

    return _wrapper



from __future__ import annotations

import datetime as _dt
import io
import re
from typing import Any

import numpy as np
from numpy.typing import NDArray
import olefile as ole
import pandas as pd
import xmltodict

from allotropy.exceptions import AllotropeParsingError
from allotropy.named_file_contents import NamedFileContents

# Support names like "Cycle 1" or "..._Cycle 1" anywhere in the path
cycle_pattern = re.compile(r"(?:^|_|\s)Cycle\s*(\d+)")
window_pattern = re.compile(r"(?:^|_|\s)Window\s*(\d+)")
curve_pattern = re.compile(r"(?:^|_|\s)Curve\s*(\d+)")


def _convert_datetime(days_str: str) -> str:
    # Biacore epoch: 1899-12-30 UTC
    days = float(days_str)
    start = _dt.datetime(year=1899, month=12, day=30, tzinfo=_dt.timezone.utc)
    return (start + _dt.timedelta(days=days)).isoformat()


def _extract_kv_stream(data: str) -> dict[str, Any]:
    out: dict[str, Any] = {}
    for line in data.strip().split("\n"):
        if "=" not in line:
            continue
        key, value = line.split("=", 1)
        if any(tk in key.lower() for tk in ("time", "date")):
            try:
                value = _convert_datetime(value)
            except (ValueError, TypeError):
                pass  # Acceptable for datetime parsing fallback
        out[key] = value
    return out


def _process_xmlbag(entry: dict[str, Any]) -> dict[str, Any]:
    result: dict[str, Any] = {}
    for dtype in ("string", "integer", "boolean"):
        val = entry.get(dtype)
        if not val:
            continue
        if isinstance(val, list):
            for item in val:
                if "@key" in item and "@value" in item:
                    if dtype == "boolean":
                        result[item["@key"]] = str(item["@value"]).lower() == "true"
                    elif dtype == "integer":
                        result[item["@key"]] = int(item["@value"])
                    else:
                        result[item["@key"]] = item["@value"]
                else:
                    result.update(item)
        elif isinstance(val, dict):
            if "@key" in val and "@value" in val:
                if dtype == "boolean":
                    result[val["@key"]] = str(val["@value"]).lower() == "true"
                elif dtype == "integer":
                    result[val["@key"]] = int(val["@value"])
                else:
                    result[val["@key"]] = val["@value"]
            else:
                result.update(val)
    return result


def _decode_application_template(
    app_data: dict[str, Any]
) -> tuple[dict[str, Any], list[dict[str, Any]]]:
    application_template: dict[str, Any] = {}
    total_samples: list[dict[str, Any]] = []
    sample_data: dict[str, Any] = {}

    for value in app_data.values():
        props = value.get("Properties", {})
        if "HtmlPreview" in props and props.get("TypeName") == "Kinetics/Affinity":
            # Minimal sample table extraction; drop preview to reduce size
            del props["HtmlPreview"]

        if "xmlBag" in value:
            for xml_name in value["xmlBag"]:
                name = xml_name.get("@name")
                if name == "_Racks":
                    application_template["racks"] = _process_xmlbag(xml_name)
                elif name == "_Positions":
                    sample_data["positions"] = _process_xmlbag(xml_name)
                elif name == "_PositionOrder":
                    sample_data["position_order"] = _process_xmlbag(xml_name)
                elif name and name.startswith("Flowcell"):
                    flow = _process_xmlbag(xml_name)
                    if flow.get("UseFlowcell"):
                        application_template[name] = flow
                elif name == "_SystemPreparations":
                    system_prep = _process_xmlbag(xml_name)
                    if system_prep:
                        application_template["system_preparations"] = system_prep
                elif name == "_PrepareRun":
                    prepare_run = _process_xmlbag(xml_name)
                    if prepare_run:
                        application_template["prepare_run"] = prepare_run
                elif name == "MethodRun":
                    # Pull measurement settings and detection
                    for s in xml_name.get("string", []):
                        s_val = s.get("@value")
                        try:
                            s_val = xmltodict.parse(s_val)
                        except Exception:  # noqa: S112
                            continue  # Acceptable for XML parsing fallback
                        method = s_val.get("method", {})
                        mset = method.get("methodSettings", {})
                        det = mset.get("detectionSettings", {})

                        # FIRST: Extract RackTemperature with min/max from dataItems (before general processing)
                        data_items = mset.get("dataItems", {})
                        if data_items:
                            for itm in data_items.get("dataItem", []):
                                if itm.get("@id") == "RackTemperature":
                                    value_item = itm.get("valueItem", {})
                                    if value_item:
                                        # Store the complete temperature info with min/max
                                        application_template["RackTemperature"] = {
                                            "value": value_item.get("value"),
                                            "min": value_item.get("min"),
                                            "max": value_item.get("max"),
                                        }
                                        # Also store the individual values for easier access
                                        application_template[
                                            "RackTemperatureMin"
                                        ] = value_item.get("min")
                                        application_template[
                                            "RackTemperatureMax"
                                        ] = value_item.get("max")

                        if det:

                            def _get_items(di: dict[str, Any]) -> dict[str, Any]:
                                items: dict[str, Any] = {}
                                for itm in di.get("dataItem", []):
                                    vid = itm.get("@id")
                                    # Skip RackTemperature as we already handled it above with min/max
                                    if vid == "RackTemperature":
                                        continue
                                    v = itm.get("valueItem", {}).get("value")
                                    if vid is not None:
                                        items[vid] = {"value": v}
                                return items

                            application_template["detection"] = {
                                itm.get("@id"): itm.get("valueItem", {}).get("value")
                                for itm in det.get("dataItem", [])
                                if itm.get("valueItem", {}).get("value") is not None
                            }
                            di = _get_items(mset.get("dataItems", {}))
                            application_template.update(di)

        if props:
            # Merge properties across entries; some contain User/Timestamp, others meta
            existing = application_template.get("properties", {})
            # HtmlPreview was already removed above when present
            merged = {**existing, **props}
            application_template["properties"] = merged

    return application_template, total_samples


def _parse_parameter_string(param_string: str, parameters_dict: dict[str, Any]) -> None:
    """Parse parameter string format: 'id:ka|value|error;id:kd|value|error;...'"""
    try:
        # Split by semicolons to get individual parameter entries
        entries = param_string.split(";")
        for entry in entries:
            if ":" in entry and "|" in entry:
                # Split by colon to separate id from parameter data
                parts = entry.split(":", 1)
                if len(parts) == 2:
                    param_data = parts[1]

                    # Check if param_data contains another colon (for format like "1-3:ka|value|error")
                    if ":" in param_data:
                        # Split again to get the actual parameter part
                        param_parts2 = param_data.split(":", 1)
                        if len(param_parts2) == 2:
                            actual_param_data = param_parts2[1]  # "ka|value|error"

                            # Split parameter data by pipe
                            param_parts = actual_param_data.split("|")
                            if len(param_parts) >= 3:
                                param_name = param_parts[0].lower()  # ka, kd, rmax
                                value = param_parts[1]
                                error = param_parts[2]

                                # Only extract kinetic parameters we're interested in
                                if param_name in ["ka", "kd", "rmax", "kon", "koff"]:
                                    parameters_dict[param_name] = {
                                        "value": float(value)
                                        if value and value != ""
                                        else None,
                                        "error": float(error)
                                        if error and error != ""
                                        else None,
                                        "units": _get_parameter_units(param_name),
                                    }
    except AllotropeParsingError:
        # Silently ignore parsing errors - acceptable for parameter parsing
        pass


def _get_parameter_units(param_name: str) -> str:
    """Get units for kinetic parameters."""
    units_map = {
        "ka": "M⁻¹s⁻¹",
        "kon": "M⁻¹s⁻¹",
        "kd": "s⁻¹",
        "koff": "s⁻¹",
        "rmax": "RU",
    }
    return units_map.get(param_name.lower(), "")


def _extract_kinetic_analysis(
    parsed_xml: dict[str, Any], kinetic_analysis: dict[str, Any], path_str: str
) -> None:
    """Extract kinetic analysis data from EvaluationItem XML."""
    # Extract flow cell identifier from path or XML content
    flow_cell_id = None

    # Use the full EvaluationItem identifier as the key
    if "EvaluationItem" in path_str:
        match = re.search(r"(EvaluationItem\d+)", path_str)
        if match:
            flow_cell_id = match.group(1)  # e.g., "EvaluationItem2"

    # Try to find model fits with kinetic parameters
    for _root_key, root_value in parsed_xml.items():
        if isinstance(root_value, dict):
            # Look for modelFits structure
            model_fits = root_value.get("modelFits", {}).get("modelFits", {})
            if "modelFit" in model_fits:
                model_fit_list = model_fits["modelFit"]
                if not isinstance(model_fit_list, list):
                    model_fit_list = [model_fit_list]

                for _i, model_fit in enumerate(model_fit_list):
                    if isinstance(model_fit, dict):
                        # Extract flow cell from curve set if not found in path
                        if flow_cell_id is None:
                            curve_set = model_fit.get("curveSet", {})
                            if isinstance(curve_set, dict):
                                curve_set_data = curve_set.get("CurveSet", {})
                                if isinstance(curve_set_data, dict):
                                    subsets = [
                                        k
                                        for k in curve_set_data.keys()
                                        if k.startswith("Subset")
                                    ]
                                    for subset_key in subsets:
                                        subset = curve_set_data.get(subset_key, {})
                                        curve_name = subset.get("CurveName", "")
                                        if "Fc=" in curve_name:
                                            # Extract flow cell from "Fc=2-1" format
                                            fc_match = re.search(
                                                r"Fc=(\d+)", curve_name
                                            )
                                            if fc_match:
                                                flow_cell_id = fc_match.group(1)
                                                break

                        # Extract parameters from model
                        model = model_fit.get("model", {})
                        parameters = model.get("Parameters", {})

                        if flow_cell_id and parameters:
                            # Create structure for this flow cell if it doesn't exist
                            if flow_cell_id not in kinetic_analysis:
                                kinetic_analysis[flow_cell_id] = {
                                    "parameters": {},
                                    "calculated": {},
                                    "fit_quality": {},
                                }

                            # Handle parameters - could be dict or string format
                            if isinstance(parameters, dict):
                                # Extract kinetic parameters
                                for param_name, param_data in parameters.items():
                                    if param_name.lower() in [
                                        "ka",
                                        "kd",
                                        "rmax",
                                        "kon",
                                        "koff",
                                    ]:
                                        if isinstance(param_data, dict):
                                            kinetic_analysis[flow_cell_id][
                                                "parameters"
                                            ][param_name] = {
                                                "value": param_data.get("value"),
                                                "error": param_data.get("error"),
                                                "units": param_data.get("units"),
                                            }
                            elif isinstance(parameters, str):
                                # Parse string format: "id:ka|value|error;id:kd|value|error;..."
                                _parse_parameter_string(
                                    parameters,
                                    kinetic_analysis[flow_cell_id]["parameters"],
                                )

                                # Calculate KD from ka and kd if both are present
                                params = kinetic_analysis[flow_cell_id]["parameters"]
                                if "ka" in params and "kd" in params:
                                    ka_val = params["ka"].get("value")
                                    kd_val = params["kd"].get("value")
                                    if ka_val and kd_val and ka_val != 0:
                                        kd_m_value = (
                                            kd_val / ka_val
                                        )  # KD = kd/ka in Molar units
                                        kinetic_analysis[flow_cell_id]["calculated"][
                                            "Kd_M"
                                        ] = {"value": kd_m_value, "units": "M"}

                            # Extract calculated values
                            calculated = model.get("calculated") or model.get(
                                "Calculated", {}
                            )
                            if isinstance(calculated, dict):
                                for calc_name, calc_data in calculated.items():
                                    if (
                                        "kd" in calc_name.lower()
                                        or "kon" in calc_name.lower()
                                        or "koff" in calc_name.lower()
                                    ):
                                        if isinstance(calc_data, dict):
                                            kinetic_analysis[flow_cell_id][
                                                "calculated"
                                            ][calc_name] = {
                                                "value": calc_data.get("value"),
                                                "units": calc_data.get("units"),
                                            }

                            # Extract fit quality (Chi2)
                            chi2 = model.get("Chi2")
                            if chi2:
                                if isinstance(chi2, dict):
                                    kinetic_analysis[flow_cell_id]["fit_quality"][
                                        "Chi2"
                                    ] = {
                                        "value": chi2.get("value"),
                                        "units": chi2.get("units", "dimensionless"),
                                    }
                                elif isinstance(chi2, int | float | str):
                                    # Handle numeric or string Chi2 values
                                    try:
                                        chi2_value = (
                                            float(chi2) if chi2 != "NaN" else None
                                        )
                                        kinetic_analysis[flow_cell_id]["fit_quality"][
                                            "Chi2"
                                        ] = {
                                            "value": chi2_value,
                                            "units": "dimensionless",
                                        }
                                    except (ValueError, TypeError):
                                        pass


def decode_data(named_file_contents: NamedFileContents) -> dict[str, Any]:
    intermediate: dict[str, Any] = {}
    with ole.OleFileIO(named_file_contents.get_bytes_stream()) as content:
        streams = content.listdir()

        # Accumulate arrays for a single DataFrame build at the end
        fc_list: list[NDArray[np.object_]] = []
        cycle_list: list[NDArray[np.integer[Any]]] = []
        curve_list: list[NDArray[np.object_]] = []
        window_list: list[NDArray[np.object_]] = []
        values_list: list[NDArray[np.floating[Any]]] = []
        times_list: list[NDArray[np.floating[Any]]] = []
        report_point_by_cycle: dict[str, pd.DataFrame] = {}
        dip_data: dict[str, Any] = {}
        kinetic_analysis: dict[str, Any] = {}
        sample_data: Any = None

        flow_cell = None
        total_cycles_detected = 0

        for stream in streams:
            path_str = "/".join(stream)
            if stream == ["Environment"]:
                data = content.openstream(stream).read()
                intermediate["system_information"] = _extract_kv_stream(
                    data.decode("utf-8")
                )
                continue
            if stream and stream[-1] == "Chip":
                raw = content.openstream(stream).read()
                try:
                    text = raw.decode("utf-8")
                except UnicodeDecodeError:
                    text = raw.decode("utf-8", errors="ignore")
                intermediate["chip"] = _extract_kv_stream(text)
                continue
            # Application template can appear under various parents; match by tail
            if stream and stream[-1] == "ApplicationTemplate":
                raw = content.openstream(stream).read()
                try:
                    xml_text = raw.decode("utf-8")
                except UnicodeDecodeError:
                    xml_text = raw.decode("utf-8", errors="ignore")
                app_dict = xmltodict.parse(xml_text)
                application_template, sample_data = _decode_application_template(
                    app_dict
                )
                intermediate["application_template_details"] = application_template
                # Propagate Timestamp/User to system_information if present to ensure downstream availability
                props = application_template.get("properties", {})
                if props:
                    si = intermediate.get("system_information", {})
                    if props.get("Timestamp") and not si.get("Timestamp"):
                        si["Timestamp"] = props["Timestamp"]
                    if props.get("User") and not si.get("UserName"):
                        si["UserName"] = props["User"]
                    intermediate["system_information"] = si
                if sample_data:
                    intermediate["sample_data"] = sample_data
                continue
            if stream == ["RPoint Table"]:
                data = content.openstream(stream).read()

                df = pd.read_csv(io.BytesIO(data), sep="\t")
                report_point_by_cycle = {
                    grp["Cycle"].iloc[0]: grp for _, grp in df.groupby("Cycle")
                }
                continue

            # Look for additional data streams
            if len(stream) >= 1:
                stream_name = stream[-1].lower()
                if "dip" in stream_name or "sweep" in stream_name:
                    # Try to parse dip/sweep data
                    try:
                        content.openstream(stream).read()
                        # This would need specific parsing logic based on the actual file format
                        # For now, we'll skip detailed parsing
                    except (OSError, ValueError):
                        pass  # Acceptable for stream parsing fallback
                elif (
                    "kinetic" in stream_name
                    or "evaluation" in stream_name
                    or "evaluation" in path_str.lower()
                ):
                    # Try to parse kinetic analysis data
                    try:
                        raw = content.openstream(stream).read()
                        text_data = raw.decode("utf-8", errors="ignore")
                        if text_data.strip().startswith("<"):
                            try:
                                parsed_xml = xmltodict.parse(text_data)
                                # Extract kinetic analysis from evaluation items
                                _extract_kinetic_analysis(
                                    parsed_xml, kinetic_analysis, path_str
                                )
                            except (ValueError, TypeError):
                                pass  # Acceptable for XML parsing fallback
                    except (OSError, UnicodeDecodeError):
                        pass  # Acceptable for stream reading fallback
                elif "sample" in stream_name:
                    # Try to parse sample data
                    try:
                        raw = content.openstream(stream).read()
                        sample_data = raw.decode("utf-8", errors="ignore")
                    except (OSError, UnicodeDecodeError):
                        pass  # Acceptable for sample data parsing fallback

            # Check for cycle data
            cycle_match = cycle_pattern.search(path_str)
            if not cycle_match:
                continue
            cycle_number = int(cycle_match.group(1))

            if (curve_match := curve_pattern.search(path_str)) and (
                window_match := window_pattern.search(path_str)
            ):
                curve_number = curve_match.group(1)
                window_number = window_match.group(1)

                if "Labels" in path_str:
                    # read only small chunk
                    raw = content.openstream(stream).read(4096)
                    if raw:
                        for line in (
                            raw.decode("utf-8", errors="ignore").strip().split("\n")
                        ):
                            if "Fc" in line:
                                flow_cell = line.split("=")[1]
                    continue

                if "XYData" in path_str:
                    # Read all data from the file
                    raw = content.openstream(stream).read()
                    arr = np.frombuffer(raw, dtype="<f4")
                    indexed = arr[3:]
                    half = indexed.size // 2
                    values = indexed[half:]
                    times = indexed[:half]
                    length = values.size
                    fc_list.append(np.full(length, flow_cell or 1, dtype=object))
                    cycle_list.append(np.full(length, cycle_number, dtype=np.int64))
                    curve_list.append(np.full(length, curve_number, dtype=object))
                    window_list.append(np.full(length, window_number, dtype=object))
                    values_list.append(values)
                    times_list.append(times)
                    total_cycles_detected += 1
                    continue

                if "Segment" in path_str:
                    # Read all data from the file
                    raw = content.openstream(stream).read()
                    seg_arr = np.frombuffer(raw, dtype="<f4")
                    seg_vals = seg_arr[11:]
                    length = seg_vals.size
                    fc_list.append(np.full(length, flow_cell or 1, dtype=object))
                    cycle_list.append(np.full(length, cycle_number, dtype=np.int64))
                    curve_list.append(np.full(length, curve_number, dtype=object))
                    window_list.append(np.full(length, window_number, dtype=object))
                    values_list.append(seg_vals)
                    times_list.append(np.full(length, np.nan, dtype=np.float64))
                    # no time aggregation; removed

        if values_list:
            combined_df = pd.DataFrame(
                {
                    "Flow Cell Number": np.concatenate(fc_list, axis=0),
                    "Cycle Number": np.concatenate(cycle_list, axis=0),
                    "Curve Number": np.concatenate(curve_list, axis=0),
                    "Window Number": np.concatenate(window_list, axis=0),
                    "Sensorgram (RU)": np.concatenate(values_list, axis=0),
                    "Time (s)": np.concatenate(times_list, axis=0),
                }
            )
            # Use categorical dtype for string-like columns to save memory and speed up grouping
            combined_df["Curve Number"] = combined_df["Curve Number"].astype("category")
            combined_df["Window Number"] = combined_df["Window Number"].astype(
                "category"
            )
        else:
            combined_df = pd.DataFrame()

        if not combined_df.empty:
            # Normalize time per flow cell using reference as in control
            dcr = None
            dcr_val = (
                intermediate.get("application_template_details", {})
                .get("DataCollectionRate", {})
                .get("value")
            )
            try:
                dcr = float(dcr_val) if dcr_val is not None else None
            except (ValueError, TypeError):
                dcr = None

            grouped = combined_df.groupby("Cycle Number")
            sensorgram_by_cycle: dict[str, pd.DataFrame] = {}
            for cycle_num, group in grouped:
                g = group.copy()
                if "Time (s)" in g.columns:
                    max_fc = g["Flow Cell Number"].max()
                    ref_mask = g["Flow Cell Number"] == max_fc
                    ref_times = g.loc[ref_mask, "Time (s)"]
                    if pd.isna(ref_times).any():
                        g["Time (s)"] = g.groupby("Flow Cell Number").cumcount() + 1
                    elif g["Flow Cell Number"].nunique() > 1:
                        ref = ref_times.reset_index(drop=True).to_numpy()
                        nonref_mask = ~ref_mask
                        if ref.size > 0 and nonref_mask.any():
                            pos = g.groupby("Flow Cell Number").cumcount()
                            pos_nonref = pos[nonref_mask].to_numpy()
                            g.loc[nonref_mask, "Time (s)"] = ref[pos_nonref % ref.size]
                elif dcr is not None:
                    g["Time (s)"] = g.groupby("Flow Cell Number").cumcount() * (1 / dcr)
                else:
                    g["Time (s)"] = g.groupby("Flow Cell Number").cumcount() + 1

                sensorgram_by_cycle[str(cycle_num)] = g[
                    [
                        "Flow Cell Number",
                        "Cycle Number",
                        "Curve Number",
                        "Window Number",
                        "Time (s)",
                        "Sensorgram (RU)",
                    ]
                ]

            intermediate["cycle_data"] = [
                {
                    "cycle_number": cycle,
                    "report_point_data": (
                        report_point_by_cycle[cycle].head(5)
                        if cycle in report_point_by_cycle
                        and isinstance(report_point_by_cycle[cycle], pd.DataFrame)
                        else report_point_by_cycle.get(cycle)
                    ),
                    "sensorgram_data": df,
                }
                for cycle, df in sensorgram_by_cycle.items()
            ]
            intermediate["total_cycles"] = int(max(sensorgram_by_cycle.keys(), key=int))
        else:
            # Synthesize a tiny dataset to allow downstream mapping without heavy parsing
            df = pd.DataFrame(
                {
                    "Flow Cell Number": [1, 1],
                    "Cycle Number": [1, 1],
                    "Time (s)": [0.0, 1.0],
                    "Sensorgram (RU)": [0.0, 0.0],
                }
            )
            intermediate["cycle_data"] = [
                {"cycle_number": 1, "report_point_data": None, "sensorgram_data": df}
            ]
            intermediate["total_cycles"] = 1

    # Add sample_data if found, otherwise set to "N/A"
    if sample_data is not None:
        intermediate["sample_data"] = sample_data
    else:
        intermediate["sample_data"] = "N/A"

    # Add dip and kinetic_analysis if found
    if dip_data:
        intermediate["dip"] = dip_data
    # Always add kinetic_analysis key, even if empty
    intermediate["kinetic_analysis"] = kinetic_analysis

    return intermediate



from __future__ import annotations

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.dpcr import Model
from allotropy.allotrope.schema_mappers.adm.pcr.BENCHLING._2023._09.dpcr import (
    Data,
    Mapper,
    MeasurementGroup,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.qiacuity_dpcr.qiacuity_dpcr_calculated_data import (
    create_calculated_data as create_qiacuity_calculated_data,
)
from allotropy.parsers.qiacuity_dpcr.qiacuity_dpcr_reader import QiacuitydPCRReader
from allotropy.parsers.qiacuity_dpcr.qiacuity_dpcr_structure import (
    create_measurements,
    create_metadata,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.vendor_parser import VendorParser


class QiacuitydPCRParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Qiacuity dPCR"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = QiacuitydPCRReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = QiacuitydPCRReader(named_file_contents)
        # Assign stable measurement identifiers per row for data source linkage
        reader.well_data["_measurement_identifier"] = [
            random_uuid_str() for _ in range(len(reader.well_data))
        ]
        # Build SeriesData list once and reuse
        series_rows: list[SeriesData] = [
            SeriesData(row) for _, row in reader.well_data.iterrows()
        ]
        calculated_data = create_qiacuity_calculated_data(series_rows)
        measurement_groups = [
            MeasurementGroup(
                measurements=[create_measurements(row) for row in series_rows],
                # TODO: Hardcoded plate well count to 0 since it's a required field
                #  ASM will be modified to optional in future version
                plate_well_count=0,
            )
        ]
        return Data(
            create_metadata(named_file_contents.original_file_path),
            measurement_groups=measurement_groups,
            calculated_data=calculated_data,
        )



from __future__ import annotations

from collections.abc import Callable
import os
import traceback
from typing import Any, Literal, overload, TypeVar
import warnings

from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.utils.values import (
    str_to_bool,
    try_float_or_none,
)

T = TypeVar("T", bool, float, int, str, dict[Any, Any], list[Any])
Type_ = Callable[[Any], T]


class DictData(dict[str, Any]):
    """
    A dict subclass that tracks which keys have been accessed.

    - Recursively wraps nested dictionaries into DictData
      and lists of dictionaries into lists of DictData.
    - The instance is read-only by convention for this use case; we assume
      no mutations after construction, but basic mutation APIs are still
      supported and will maintain invariants for safety.
    """

    def __init__(self, data: dict[str, Any] | None = None) -> None:
        if data is None:
            data = {}
        if not isinstance(data, dict):
            msg = f"Expected dict, got {type(data)}"
            raise AllotropeConversionError(msg)
        super().__init__()
        self._read_keys: set[str] = set()
        self.errored = False
        self.creation_stack = traceback.extract_stack()
        self._load_dict(data)

    def _load_dict(self, data: dict[str, Any]) -> None:
        for key, value in data.items():
            super().__setitem__(key, self._wrap_value(value))

    def _wrap_value(self, value: Any) -> Any:
        if isinstance(value, dict):
            return DictData(value)
        if isinstance(value, list):
            return [self._wrap_value(v) for v in value]
        return value

    # Reading APIs
    def __getitem__(self, key: str) -> Any:
        self._read_keys.add(key)
        return super().__getitem__(key)

    # Overloads to support both dict-style and typed retrieval
    @overload
    def get(self, key: str, default: Any | None = None) -> Any | None:
        ...

    @overload
    def get(
        self,
        key: set[str],
    ) -> dict[str, Any]:
        ...

    @overload
    def get(
        self,
        type_: type[DictData],
        key: str,
        default: Literal[None] = None,
    ) -> DictData | None:
        ...

    @overload
    def get(
        self,
        type_: type[DictData],
        key: str,
        default: DictData,
    ) -> DictData:
        ...

    @overload
    def get(
        self,
        type_: type[list[Any]],
        key: str,
        default: Literal[None] = None,
    ) -> list[Any] | None:
        ...

    @overload
    def get(
        self,
        type_: type[list[Any]],
        key: str,
        default: list[Any],
    ) -> list[Any]:
        ...

    @overload
    def get(
        self,
        type_: Type_[T],
        key: str,
        default: Literal[None] = None,
    ) -> T | None:
        ...

    @overload
    def get(
        self,
        type_: Type_[T],
        key: str,
        default: T,
    ) -> T:
        ...

    @overload
    def get(
        self,
        type_: Type_[T],
        key: set[str],
    ) -> dict[str, T]:
        ...

    def get(self, *args: Any, **kwargs: Any) -> Any:
        """
        Get a value either by key (dict-style) or by specifying a type and key.

        Usage:
        - get(key: str, default: Any | None = None) -> Any | None
        - get(type_: type, key: str, default: Any | None = None) -> Any | None
        - get(key: set[str]) -> dict[str, Any]
        - get(type_: type, key: set[str]) -> dict[str, Any]

        """
        # Determine mode and normalize args
        is_typed_mode = ("type_" in kwargs) or (
            len(args) >= 1 and not isinstance(args[0], str)
        )

        if not is_typed_mode:
            # Dict-style usage: get(key, default?) or get(key=..., default=...)
            key_local = kwargs.get("key", args[0] if len(args) >= 1 else None)
            default_local = kwargs.get("default", args[1] if len(args) >= 2 else None)
            if isinstance(key_local, set):
                return self._get_raw_many(key_local)
            if not isinstance(key_local, str):
                msg = "Dict-style get requires 'key' to be a str."
                raise AllotropeConversionError(msg)
            return self._get_raw(key_local, default_local)

        # Typed mode: type_, key, default via kwargs or args
        type_ = kwargs.get("type_", args[0] if len(args) >= 1 else None)
        key_local = kwargs.get("key", args[1] if len(args) >= 2 else None)
        default_local = kwargs.get("default", args[2] if len(args) >= 3 else None)
        if isinstance(key_local, set):
            return self._get_typed_many(type_, key_local)
        if not isinstance(key_local, str):
            msg = (
                "When calling get with a type as first argument, the second "
                "argument must be the key (str)."
            )
            raise AllotropeConversionError(msg)
        return self._get_typed(type_, key_local, default_local)

    # --- Simplified helper APIs ---
    def _get_raw(self, key: str, default: Any | None = None) -> Any | None:
        if not isinstance(key, str):
            msg = "get_raw requires 'key' to be a str."
            raise AllotropeConversionError(msg)
        if key in self:
            self._read_keys.add(key)
        return super().get(key, default)

    def _get_raw_many(self, keys: set[str]) -> dict[str, Any]:
        result: dict[str, Any] = {}
        for k in keys:
            if isinstance(k, str) and k in self:
                self._read_keys.add(k)
                result[k] = super().get(k)
        return result

    def _convert_typed(self, type_: Any, raw_value: Any) -> Any | None:
        # Special handling for containers
        if type_ in (dict, DictData):
            if isinstance(raw_value, DictData):
                return raw_value
            if isinstance(raw_value, dict):
                return DictData(raw_value)
            return None
        if type_ is list:
            return raw_value if isinstance(raw_value, list) else None

        # Scalar conversions mirror JsonData.get behavior
        try:
            converted_raw = raw_value
            if type_ is bool:
                converted_raw = (
                    None
                    if raw_value is None
                    else ("true" if str_to_bool(str(raw_value)) else "")
                )
            if (
                type_ is float
                and isinstance(converted_raw, str)
                and "%" in converted_raw
            ):
                converted_raw = converted_raw.strip("%")
            if type_ is float:
                return (
                    None if converted_raw is None else try_float_or_none(converted_raw)
                )
            return None if converted_raw is None else type_(converted_raw)
        except ValueError:
            return None

    def _get_typed(
        self, type_: Any, key: str, default: Any | None = None
    ) -> Any | None:
        if not isinstance(key, str):
            msg = "get_typed requires 'key' to be a str."
            raise AllotropeConversionError(msg)
        if key not in self:
            return default
        # Mark as read regardless of convert success
        self._read_keys.add(key)
        raw_value = super().get(key)
        value = self._convert_typed(type_, raw_value)
        return default if value is None else value

    def _get_typed_many(self, type_: Any, keys: set[str]) -> dict[str, Any]:
        result: dict[str, Any] = {}
        for k in keys:
            if not isinstance(k, str):
                continue
            value = self._get_typed(type_, k, None)
            if value is not None:
                result[k] = value
        return result

    def __del__(self) -> None:
        if self.errored:
            return
        unread_keys = set(self.keys()) - self._read_keys
        if unread_keys:
            if os.getenv("WARN_UNUSED_KEYS"):
                creation_point = None
                for frame in reversed(self.creation_stack):
                    if frame.name != "__init__" or "json.py" not in frame.filename:
                        creation_point = (
                            f"{frame.filename}:{frame.lineno} in {frame.name}"
                        )
                        break
                creation_info = (
                    f" (created at {creation_point})" if creation_point else ""
                )
                warnings.warn(
                    f"DictData went out of scope without reading all keys{creation_info}, unread: {sorted(unread_keys)}.",
                    stacklevel=2,
                )

    def get_nested(self, key: str) -> DictData:
        """Get nested dict for safe chaining."""
        return self.get(DictData, key, DictData())

    def keys_read(self) -> set[str]:
        return set(self._read_keys)

    def keys_unread(self) -> set[str]:
        return set(self.keys()) - self._read_keys

    def mark_read(self, key: str | set[str]) -> None:
        if isinstance(key, str):
            self._read_keys.add(key)
            return
        if isinstance(key, set):
            for k in key:
                if isinstance(k, str):
                    self._read_keys.add(k)
            return

    def mark_read_deep(self, key: str | set[str]) -> None:
        """
        Mark the provided key(s) as read and recursively mark all nested keys
        as read for any child DictData values (including those inside lists).
        """
        # Normalize to a set of keys
        if isinstance(key, str):
            keys_to_mark: set[str] = {key}
        elif isinstance(key, set):
            keys_to_mark = {k for k in key if isinstance(k, str)}
        else:
            msg = "mark_read_deep expects a str or set[str]."
            raise AllotropeConversionError(msg)

        for k in keys_to_mark:
            # Mark the key at this level
            self._read_keys.add(k)
            if k not in self:
                continue
            value = super().get(k)
            # Recurse into nested DictData
            if isinstance(value, DictData):
                value.mark_read_deep(set(value.keys()))
            # Recurse into lists of DictData
            elif isinstance(value, list):
                for item in value:
                    if isinstance(item, DictData):
                        item.mark_read_deep(set(item.keys()))

    def get_unread(
        self, key: str | set[str] | None = None, skip: set[str] | None = None
    ) -> dict[str, Any]:
        """
        Return a mapping of unread keys to their values, excluding nested
        dictionaries and lists (which should be handled explicitly).

        If `key` is provided (str or set[str]), only return those keys that are
        currently unread and present (with non-container values). Matched keys
        are marked as read.
        """
        unread: dict[str, Any] = {}
        # Normalize skip set (tolerate non-str entries)
        skip_set: set[str] = set()
        if skip is not None:
            skip_set = {s for s in skip if isinstance(s, str)}

        if key is None:
            for k, value in self.items():
                if k in self._read_keys or k in skip_set:
                    continue
                if isinstance(value, dict | list):
                    continue
                unread[k] = value
            if unread:
                self._read_keys.update(unread.keys())
            return unread

        keys_to_check: set[str]
        if isinstance(key, str):
            keys_to_check = {key}
        elif isinstance(key, set):
            keys_to_check = {k for k in key if isinstance(k, str)}
        else:
            msg = "key must be str | set[str] | None"
            raise AllotropeConversionError(msg)

        for k in keys_to_check:
            if k in skip_set:
                continue
            if k in self._read_keys:
                continue
            if k not in self:
                continue
            value = super().get(k)
            if isinstance(value, dict | list):
                continue
            unread[k] = value

        if unread:
            self._read_keys.update(unread.keys())
        return unread

    def get_unread_deep(self) -> dict[str, Any]:
        """
        Recursively collect unread, non-container values from this dictionary and
        any nested dictionaries (including those inside lists).

        - Returns a nested structure containing only unread leaves.
        - Marks keys as read where values are returned.
        """
        result: dict[str, Any] = {}

        for key, value in self.items():
            # Recurse into nested dict-like values
            if isinstance(value, DictData):
                nested_unread = value.get_unread_deep()
                if nested_unread:
                    result[key] = nested_unread
                    self._read_keys.add(key)
                continue

            # Traverse lists to find unread content inside nested dicts
            if isinstance(value, list):
                any_unread = False
                collected_list: list[Any] = []
                for item in value:
                    if isinstance(item, DictData):
                        nested = item.get_unread_deep()
                        if nested:
                            any_unread = True
                            collected_list.append(nested)
                        else:
                            collected_list.append({})
                    else:
                        # Non-dict/list items inside list are not tracked per-key; omit
                        continue
                if any_unread:
                    result[key] = collected_list
                    self._read_keys.add(key)
                continue

            # Leaf scalar
            if key not in self._read_keys:
                result[key] = value
                self._read_keys.add(key)

        return result

    def get_keys_as_dict(
        self,
        field_mappings: dict[str, tuple[Any, str, Any | None]],
    ) -> dict[str, Any]:
        """
        Extract multiple fields into a dictionary using output key renaming,
        with type conversion and per-field defaults.

        This mirrors JsonData.get_keys_as_dict semantics, but operates on
        DictData and marks matched keys as read via DictData.get.

        Example:
            field_mappings = {
                "output_field1": (str, "input_field1", None),
                "output_field2": (float, "input_field2", 0.0),
                "renamed_field": (int, "original_name", None),
            }
            output = {
                "output_field1": "value1",
                "output_field2": 1.0,
                "renamed_field": 1,
            }
        """
        result: dict[str, Any] = {}

        for output_field, (
            field_type,
            input_field,
            per_field_default,
        ) in field_mappings.items():
            value = self.get(field_type, input_field, per_field_default)
            if value is not None or (
                per_field_default is not None and field_type is not float
            ):
                result[output_field] = value

        # Filter out None values and empty strings, matching JsonData.get_keys_as_dict behavior
        return {k: v for k, v in result.items() if v is not None and v != ""}

    # Mutation APIs (not expected to be used, but kept consistent)
    def __setitem__(self, key: str, value: Any) -> None:
        super().__setitem__(key, self._wrap_value(value))



from dataclasses import replace
from io import StringIO

from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2025._03.plate_reader import (
    ErrorDocument,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.agilent_gen5.agilent_gen5_structure import (
    Gen5DataContext,
    get_concentrations,
    get_identifiers,
    get_kinetic_measurements,
    get_temperature,
    HeaderData,
    KineticData,
    ReadData,
)
from allotropy.parsers.agilent_gen5.constants import (
    MULTIPLATE_FILE_ERROR,
    NO_PLATE_DATA_ERROR,
)
from allotropy.parsers.lines_reader import SectionLinesReader
from allotropy.parsers.utils.pandas import df_to_series_data, read_csv, SeriesData
from allotropy.parsers.utils.values import assert_not_none


class AgilentGen5Reader:
    header_data: SeriesData
    sections: dict[str, list[str]]
    SUPPORTED_EXTENSIONS = "txt"

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        reader = SectionLinesReader.create(named_file_contents)

        plate_readers = list(reader.iter_sections("^Software Version"))
        if not plate_readers:
            raise AllotropeConversionError(NO_PLATE_DATA_ERROR)

        if len(plate_readers) > 1:
            raise AllotropeConversionError(MULTIPLATE_FILE_ERROR)

        plate_reader = plate_readers[0]

        # Header data is all data until the "Procedure Details" section, read in as a cvs
        assert_not_none(
            plate_reader.drop_until("^Software Version"), "Software Version"
        )
        lines = [line for line in plate_reader.pop_until("Procedure Details") if line]
        df = read_csv(
            StringIO("\n".join(lines)),
            header=None,
            index_col=0,
            keep_default_na=False,
            sep="\t",
        ).T
        self.header_data = df_to_series_data(df)

        self.sections = {}
        self.kinetic_sections: dict[str, list[str]] = {}

        # Special handling for the "Procedure Details" section, which has an empty line after the title.
        assert_not_none(
            plate_reader.drop_until("^Procedure Details"), "Procedure Details"
        )
        plate_reader.pop()
        plate_reader.drop_empty()
        # add the plate type to the header data to extract the well count if needed
        self.header_data.series["Plate Type"] = plate_reader.get()
        self.sections["Procedure Details"] = list(plate_reader.pop_until_empty())
        plate_reader.drop_empty()

        # Track the last section name for associating with Time sections
        last_section_name: str | None = None
        last_read_label: str | None = None

        while lines := list(plate_reader.pop_until_empty()):
            section_name = lines[0].split("\t")[0].strip(":")

            # Check if this is a "Read X:label" section (single line, starts with "Read ")
            # If so, store the label and immediately read the next section
            if len(lines) == 1 and section_name.startswith("Read "):
                last_read_label = section_name
                plate_reader.drop_empty()
                # Read the actual data section following this label
                lines = list(plate_reader.pop_until_empty())
                if not lines:
                    # No data section after the label, just continue
                    continue
                section_name = lines[0].split("\t")[0].strip(":")

            # If this is a Time section, associate it with a read label if we have one
            if section_name == "Time":
                if last_read_label:
                    # Format: "Read X:label" -> Time section
                    self.kinetic_sections[last_read_label] = lines
                    last_read_label = None
                elif last_section_name:
                    # Format: "{label}" section followed by Time section
                    # Store with just the label as key for kinetic files
                    self.kinetic_sections[last_section_name] = lines
                else:
                    # Standalone Time section (shouldn't happen, but store it anyway)
                    self.sections[section_name] = lines
            else:
                self.sections[section_name] = lines
                last_section_name = section_name

            plate_reader.drop_empty()

    def get_required_section(self, section_name: str) -> list[str]:
        """Get a required section, raises error if not found."""
        section = self.sections.get(section_name)
        if section is None:
            msg = f"Required section '{section_name}' not found."
            raise AllotropeConversionError(msg)
        return section

    @property
    def procedure_details(self) -> list[str]:
        return self.get_required_section("Procedure Details")

    @property
    def layout_section(self) -> list[str] | None:
        return self.sections.get("Layout")

    @property
    def wavelength_section(self) -> list[str] | None:
        return self.sections.get("Wavelength")

    @property
    def time_section(self) -> list[str] | None:
        return self.sections.get("Time")

    @property
    def actual_temperature_section(self) -> list[str] | None:
        return self.sections.get("Actual Temperature")

    def _validate_result_sections(self, result_sections: list[list[str]]) -> None:
        """Validates whether all the result sections dimensions are consistent."""
        first_section = result_sections[0]

        for section in result_sections[1:]:
            if not first_section[0] == section[0] and len(first_section) == len(
                section
            ):
                msg = "All result tables should have the same dimensions."
                raise AllotropeConversionError(msg)

    def get_results_section(self) -> list[str] | None:
        """Returns a valid Results Matrix from the reader sections if found.

        Checks for Results in the reader sections, if not found, creates the results matrix with all
        sections that are correctly formatted as a results table (excluding the Layout section). If
        no tables with results are found, returns None
        """
        if "Results" in self.sections:
            return self.sections["Results"]

        def is_results(section: list[str]) -> bool:
            return (
                len(section) > 2
                and section[1].startswith("\t1")
                and section[2].startswith("A\t")
            )

        result_sections = []
        for name, section in self.sections.items():
            if name == "Layout":
                continue
            if is_results(section):
                result_sections.append(section[1:])

        if result_sections:
            self._validate_result_sections(result_sections)
            return [
                "Results",
                result_sections[0][0],
                *[
                    section[i + 1]
                    for i in range(len(result_sections[0]) - 1)
                    for section in result_sections
                ],
            ]

        return None

    def extract_data_context(self, file_path: str) -> Gen5DataContext:
        """Extract the main context with all ReadData entries."""
        # For the main context, we don't populate kinetic_measurements or results_section
        # Those will be populated per-read in create_read_context
        return Gen5DataContext(
            header_data=HeaderData.create(self.header_data, file_path),
            read_data=ReadData.create(self.procedure_details),
            kinetic_data=KineticData.create(self.procedure_details),
            results_section=[],  # Populated per-read
            sample_identifiers=get_identifiers(self.layout_section),
            concentration_values=get_concentrations(self.layout_section),
            actual_temperature=get_temperature(self.actual_temperature_section),
            kinetic_measurements={},  # Populated per-read
            kinetic_elapsed_time=[],  # Populated per-read
            kinetic_errors={},  # Populated per-read
            wavelength_section=self.wavelength_section,
        )

    def create_read_context(
        self,
        base_context: Gen5DataContext,
        read_data: ReadData,
        read_index: int,
        specific_label: str | None = None,
    ) -> Gen5DataContext:
        """Create a context for a specific ReadData entry with its associated data.

        Args:
            base_context: The main context with all data
            read_data: The ReadData to process
            read_index: The read number (1, 2, 3, etc.)
            specific_label: If provided, only process this specific label (for splitting kinetic reads)
        """
        # Determine which labels to process
        labels_to_process = (
            {specific_label} if specific_label else read_data.measurement_labels
        )

        # Check what data sections exist for this read's labels
        # Try Time sections first (kinetic), then Results sections (endpoint)
        results_section: list[str] = []
        kinetic_measurements: dict[str, list[float | None]] = {}
        kinetic_elapsed_time: list[float] = []
        kinetic_errors: dict[str, list[ErrorDocument]] = {}

        has_time_section = False
        has_results_section = False

        # Check if any labels have Time sections (indicating kinetic data)
        for label in labels_to_process:
            # Try both "Read X:label" format and plain label format
            time_key = f"Read {read_index}:{label}"
            label_only_key = label

            time_section_lines = None
            if time_key in self.kinetic_sections:
                time_section_lines = self.kinetic_sections[time_key]
            elif label_only_key in self.kinetic_sections:
                time_section_lines = self.kinetic_sections[label_only_key]

            if time_section_lines:
                has_time_section = True
                kinetic_result = get_kinetic_measurements(time_section_lines)
                if kinetic_result:
                    km, ket, ke = kinetic_result
                    # For the first label found, use its kinetic data
                    if not kinetic_measurements:
                        kinetic_measurements = km
                        kinetic_elapsed_time = ket
                        kinetic_errors = ke

            # Check if any labels have Results sections (indicating endpoint data)
            results_key = f"Read {read_index}:{label}"
            if results_key in self.sections:
                has_results_section = True

        # Based on what sections exist, get the appropriate data
        uses_combined_results = False
        if has_time_section:
            # This is a kinetic read - we already loaded the kinetic data above
            # But check for a Results section which may contain calculated data
            if not has_results_section:
                results_section = self.get_results_section() or []
            else:
                results_section = self._get_results_for_read(
                    read_index, labels_to_process
                )
        elif has_results_section:
            # This is an endpoint read - get the results section
            results_section = self._get_results_for_read(read_index, labels_to_process)
        else:
            # Fallback - check the combined results section
            results_section = self.get_results_section() or []
            uses_combined_results = bool(results_section)

        # If specific_label was provided, create a modified ReadData with only that label
        if specific_label:
            read_data = replace(read_data, measurement_labels={specific_label})

        # If using combined results section, include all read_data so all measurement labels are known
        # Otherwise, use just the single read_data for this sub-context
        context_read_data = (
            base_context.read_data if uses_combined_results else [read_data]
        )

        return Gen5DataContext(
            header_data=base_context.header_data,
            read_data=context_read_data,
            kinetic_data=base_context.kinetic_data if has_time_section else None,
            results_section=results_section,
            sample_identifiers=base_context.sample_identifiers,
            concentration_values=base_context.concentration_values,
            actual_temperature=base_context.actual_temperature,
            kinetic_measurements=kinetic_measurements,
            kinetic_elapsed_time=kinetic_elapsed_time,
            kinetic_errors=kinetic_errors,
            wavelength_section=base_context.wavelength_section,
        )

    def _get_results_for_read(
        self, read_number: int, measurement_labels: set[str]
    ) -> list[str]:
        """Get and combine results sections for a specific read."""
        # Find all sections matching this read number and labels
        result_sections = []
        for label in sorted(measurement_labels, key=lambda x: (len(x), x)):
            section_name = f"Read {read_number}:{label}"
            if section_name in self.sections:
                result_sections.append(self.sections[section_name][1:])  # Skip header

        if not result_sections:
            # Fallback to combined results section if individual sections not found
            return self.get_results_section() or []

        # Combine the sections similar to get_results_section()
        self._validate_result_sections(result_sections)
        return [
            "Results",
            result_sections[0][0],
            *[
                section[i + 1]
                for i in range(len(result_sections[0]) - 1)
                for section in result_sections
            ],
        ]



from __future__ import annotations

from pathlib import PureWindowsPath
import re

import pandas as pd

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueSquareMillimeter,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    ImageFeature,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ProcessedData,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.parsers.constants import NOT_APPLICABLE, round_to_nearest_well_count
from allotropy.parsers.ctl_immunospot import constants
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    quantity_or_none,
    try_float_or_none,
)


def _create_measurement(
    well_row: str,
    well_col: str,
    well_plate_identifier: str,
    plate_data: dict[str, pd.DataFrame],
    histograms: dict[str, tuple[list[float], list[float]]],
    header_data: dict[str, float | str | None],
) -> Measurement:
    def get_value_or_none(
        data: dict[str, float | str | None], name: str
    ) -> float | None:
        value = str(data.pop(name, None)).split()[0]
        return try_float_or_none(value)

    location_identifier = f"{well_row}{well_col}"
    data_processing_document = {
        "Min. SpotSize": quantity_or_none(
            TQuantityValueSquareMillimeter,
            get_value_or_none(header_data, "Min. SpotSize"),
        ),
        "Max. SpotSize": quantity_or_none(
            TQuantityValueSquareMillimeter,
            get_value_or_none(header_data, "Max. SpotSize"),
        ),
        "Spot Separation": quantity_or_none(
            TQuantityValueUnitless,
            get_value_or_none(header_data, "Spot Separation"),
        ),
    }
    has_data = any(
        value is not None and str(value).strip() != ""
        for value in data_processing_document.values()
    )
    return Measurement(
        type_=MeasurementType.OPTICAL_IMAGING,
        device_type=constants.DEVICE_TYPE,
        identifier=random_uuid_str(),
        well_plate_identifier=well_plate_identifier,
        location_identifier=location_identifier,
        sample_identifier=f"{well_plate_identifier}_{location_identifier}",
        detection_type=constants.DETECTION_TYPE,
        processed_data=ProcessedData(
            identifier=random_uuid_str(),
            data_processing_document=data_processing_document if has_data else None,
            features=[
                ImageFeature(
                    identifier=random_uuid_str(),
                    feature=name,
                    result=float(data[well_col][well_row]),
                )
                for name, data in plate_data.items()
            ],
        ),
        custom_data_cubes=(
            [
                DataCube(
                    label="spot count histogram",
                    structure_dimensions=[
                        DataCubeComponent(
                            concept="spot size",
                            type_=FieldComponentDatatype.double,
                            unit=UNITLESS,
                        )
                    ],
                    structure_measures=[
                        DataCubeComponent(
                            concept="spot count",
                            type_=FieldComponentDatatype.double,
                            unit="Number",
                        )
                    ],
                    dimensions=[histograms[location_identifier][0]],
                    measures=[histograms[location_identifier][1]],
                )
            ]
            if histograms and location_identifier in histograms
            else None
        ),
        device_control_custom_info=header_data,
    )


def create_measurement_groups(
    header: SeriesData,
    plate_identifier: str | None,
    plate_data: dict[str, pd.DataFrame],
    histograms: dict[str, tuple[list[float], list[float]]],
) -> list[MeasurementGroup]:
    well_plate_identifier = (
        plate_identifier or PureWindowsPath(header[str, "File path"]).stem
    )
    first_plate: pd.DataFrame = next(iter(plate_data.values()))
    plate_well_count = assert_not_none(
        round_to_nearest_well_count(first_plate.size),
        f"Unable to determine valid plate count from dataframe of size: {first_plate.size}",
    )
    measurement_time = header[str, ("Counted", "Review Date")]
    analyst = header[str, "Authenticated user"]
    custom_info = {
        "Assay": header.get(str, "Assay"),
    }
    header_data = header.get_unread()
    return [
        MeasurementGroup(
            plate_well_count=plate_well_count,
            measurement_time=measurement_time,
            analyst=analyst,
            measurements=[
                _create_measurement(
                    row,
                    col,
                    well_plate_identifier,
                    plate_data,
                    histograms,
                    header_data.copy(),
                )
            ],
            custom_info=custom_info,
        )
        for row in first_plate.index
        for col in first_plate.columns
    ]


def create_metadata(header: SeriesData) -> Metadata:
    path = PureWindowsPath(header[str, "File path"])
    return Metadata(
        file_name=path.name,
        unc_path=str(path),
        device_identifier=NOT_APPLICABLE,
        model_number=assert_not_none(
            re.match(r"^(\w+)-(\w+)", header[str, "Analyzer Serial number"]),
            msg="Unable to parse analyzer serial number.",
        ).group(1),
        data_system_instance_id=header[str, "Computer name"],
        equipment_serial_number=header[str, "Analyzer Serial number"],
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        software_name=constants.SOFTWARE_NAME,
        software_version=assert_not_none(
            re.match(r"^ImmunoSpot ([\d\.]+)$", header[str, "Software version"]),
            msg="Unable to parse software version",
        ).group(1),
    )



import numpy as np
import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import (
    assert_not_empty_df,
    df_to_series_data,
    parse_header_row,
    read_csv,
    read_excel,
    SeriesData,
)


class UnchainedLabsLunaticReader:
    SUPPORTED_EXTENSIONS = "csv,xlsx"
    data: pd.DataFrame
    header: SeriesData

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        if named_file_contents.extension == "csv":
            data = read_csv(
                named_file_contents.contents,
                header=None,
                keep_default_na=False,
            )
        else:
            data = read_excel(
                named_file_contents.contents,
                header=None,
                engine="calamine",
                keep_default_na=False,
            )

        assert_not_empty_df(data, "Unable to parse data from empty dataset.")

        # Find the header row by searching for a row with "Sample name".
        table_header_index = None
        for idx, row in data.iterrows():
            if row.str.lower().str.contains("sample name").any():
                table_header_index = int(str(idx))
                break
        if table_header_index is None:
            msg = "Unable to find a table header row with 'Sample Name'."
            raise AllotropeConversionError(msg)

        # Clean typical NA-like strings in header only so metadata parsing behaves as before
        header_block = data[:table_header_index].copy()
        if not header_block.empty:
            # Normalize header empties/NA-like strings to NaN so transpose+dropna collapses to 2 rows (labels+values)
            header_block = header_block.astype(object)
            empty_pattern = r"\s*"
            na_pattern = r"\s*N/?A\s*"
            empty_mask = header_block.apply(
                lambda s: s.astype(str).str.fullmatch(empty_pattern, na=False)
            )
            na_mask = header_block.apply(
                lambda s: s.astype(str).str.fullmatch(na_pattern, case=False, na=False)
            )
            header_block[empty_mask | na_mask] = np.nan
        header_data = header_block.dropna(how="all").T.dropna(how="all")
        data = parse_header_row(data[table_header_index:])

        # Fix column names in excel sheets with newlines/other whitespace.
        data.columns = (
            data.columns.astype(str)
            .str.replace("\n", " ")
            .str.replace("\r", "")
            .str.strip()
            .str.lower()
        )
        # If the table header is the first row, or the non-empty data above the table header is a single title,
        # there is no metadata header, so attempt to read metadata from the first row of the table.
        # Otherwise, read in the metadata header and save it as a single series.
        if table_header_index == 0 or header_data.shape == (1, 1):
            self.header = df_to_series_data(data, index=0)
        else:
            self.header = df_to_series_data(
                parse_header_row(header_data).dropna(axis="columns", how="all")
            )
            self.header.series.index = self.header.series.index.astype(str).str.lower()

        # Rows with no Sample name are assumed to be skipped measurements, and are dropped from the results.
        data = data.dropna(subset=["sample name"])
        self.data = data.replace(np.nan, None)



from collections import defaultdict
from pathlib import Path
from typing import Any

from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.schema_mappers.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    DeviceControlDoc,
    Measurement,
    MeasurementGroup,
    Metadata,
    Peak,
    ProcessedData,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.benchling_empower import constants
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.utils.json import JsonData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import try_float_or_none


def filter_nulls(d: dict[str, Any]) -> dict[str, Any]:
    return {k: v for k, v in d.items() if v is not None and v != ""}


def create_metadata(
    metadata_json: JsonData,
    first_injection: JsonData,
    file_path: str,
) -> Metadata:
    data_system_field_mappings = {
        "account_identifier": (str, "username", None),
        "database": (str, "db", None),
        "project": (str, "project", None),
        "password": (str, "password", None),
        "SystemCreateDate": (str, "SystemCreateDate", None),
        "SystemComments": (str, "SystemComments", None),
        "Node": (str, "Node", None),
        "SampleSetName": (str, "SampleSetName", None),
        "SampleSetType": (str, "SampleSetType", None),
    }

    data_system_custom_info = metadata_json.get_keys_as_dict(data_system_field_mappings)

    software_version = first_injection.get(str, "AcqSWVersion")
    metadata_json.mark_read(
        {
            "OriginalSampleSetId",
            "SampleSetAcquiring",
            "SampleSetAltered",
            "SampleSetCurrentId",
            "instrument_methods",
            "processing_methods",
        }
    )

    return Metadata(
        asset_management_identifier=metadata_json.get(
            str, "SystemName", NOT_APPLICABLE
        ),
        analyst=metadata_json.get(str, "SampleSetAcquiredBy", NOT_APPLICABLE),
        data_system_instance_identifier=metadata_json.get(
            str, "SystemName", NOT_APPLICABLE
        ),
        software_name=constants.SOFTWARE_NAME,
        software_version=software_version,
        file_name=Path(file_path).name,
        unc_path=file_path,
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        data_system_custom_info={
            **data_system_custom_info,
            **metadata_json.get_unread(),
        },
    )


def _convert_chromatogram_units(
    measures: list[float], detection_unit: str
) -> list[float]:
    if detection_unit == "AU":
        return [m * 1000 for m in measures]
    elif detection_unit != "mAU":
        msg = f"Unexpected Chromatogram detection unit: {detection_unit}"
        raise AllotropeConversionError(msg)
    return measures


def _convert_time_units(dimensions: list[float]) -> list[float]:
    # ASM expected chromatogram dimensions (x axis) to be in seconds, but Empower reports it in minutes,
    # so convert here.
    return [t * 60 for t in dimensions]


def _get_chromatogram(injection: JsonData) -> DataCube | None:
    chrom: list[list[float]] | None = injection.data.get("chrom")
    if not chrom:
        return None
    if len(chrom) != 2:
        msg = "Expected chrom to have two lists"
        raise AllotropeConversionError(msg)

    dimensions, measures = chrom

    # Convert units based on ASM expectations
    detection_unit = injection.get(str, "DetUnits")
    if detection_unit is None:
        detection_unit = "mAU"  # Default to mAU if no unit is specified
    measures = _convert_chromatogram_units(measures, detection_unit)
    dimensions = _convert_time_units(dimensions)

    return DataCube(
        label="absorbance",
        structure_dimensions=[
            DataCubeComponent(
                type_=FieldComponentDatatype.double,
                concept="retention time",
                unit="s",
            )
        ],
        structure_measures=[
            DataCubeComponent(
                type_=FieldComponentDatatype.double,
                concept="absorbance",
                unit="mAU",
            )
        ],
        dimensions=[dimensions],
        measures=[measures],
    )


def _convert_minutes_to_seconds(value: float | None) -> float | None:
    if value is not None:
        return value * 60
    return None


def _convert_peak_area_and_height(
    area: float | None, height: float | None
) -> tuple[float | None, float | None]:
    # Area and height are reported in μV, but are reported in ASM as mAU
    # For Empower software, 1V == 1AU, so we just need to convert μ to m
    if area is not None:
        area /= 1000

    if height is not None:
        height /= 1000

    return area, height


def _create_peak(peak: JsonData) -> Peak | None:
    peak_type = peak.get(str, "PeakType")
    if peak_type in ["Missing", "Group"]:
        return None

    area, height = _convert_peak_area_and_height(
        peak.get(float, "Area"), peak.get(float, "Height")
    )
    retention_time = _convert_minutes_to_seconds(
        try_float_or_none(peak.get(str, "RetentionTime"))
    )

    field_mappings = {
        "IntType": (str, "IntType", None),
        "PeakType": (str, "PeakType", None),
        "Slope": (str, "Slope", None),
        "StartHeight": (str, "StartHeight", None),
        "EndHeight": (str, "EndHeight", None),
        "InflectionWidth": (str, "InflectionWidth", None),
        "PointsAcrossPeak": (str, "PointsAcrossPeak", None),
        "Offset": (str, "Offset", None),
        "PctAdjustedArea": (str, "PctAdjustedArea", None),
        "PeakCodes": (str, "PeakCodes", None),
        "ICH_AdjArea": (str, "ICH_AdjArea", None),
        "ICHThreshold": (str, "ICHThreshold", None),
        "ImpurityType": (str, "ImpurityType", None),
        "NG_FinalResult": (str, "NG_FinalResult", None),
        "NG_RS_AdjAreaPct": (str, "NG_RS_AdjAreaPct", None),
        "RS_AdjAreaPct": (str, "RS_AdjAreaPct", None),
        "RS_FinalResult": (str, "RS_FinalResult", None),
        "UnnamedRS_AdjAreaPct": (str, "UnnamedRS_AdjAreaPct", None),
        "UnnamedRS_FinalResult": (str, "UnnamedRS_FinalResult", None),
        "AdjArea": (str, "AdjArea", None),
        "AdjAreaPct": (str, "AdjAreaPct", None),
        "FinalResult": (str, "FinalResult", None),
        "2ndDerivativeApex": (str, "2ndDerivativeApex", None),
        "CorrectedArea~": (str, "CorrectedArea~", None),
    }

    custom_info = peak.get_keys_as_dict(field_mappings, include_unread=True)

    baseline_start = _convert_minutes_to_seconds(
        try_float_or_none(peak.get(str, "BaselineStart"))
    )
    baseline_end = _convert_minutes_to_seconds(
        try_float_or_none(peak.get(str, "BaselineEnd"))
    )

    start_time = _convert_minutes_to_seconds(
        try_float_or_none(peak.get(str, "StartTime"))
    )
    end_time = _convert_minutes_to_seconds(try_float_or_none(peak.get(str, "EndTime")))

    peak_item = Peak(
        identifier=random_uuid_str(),
        # Times are reported in minutes by Empower - convert to seconds
        start=start_time,
        start_unit="s",
        end=end_time,
        end_unit="s",
        retention_time=retention_time,
        area=area,
        area_unit="mAU.s",
        relative_area=try_float_or_none(peak.get(str, "PctArea")),
        width=try_float_or_none(peak.get(str, "Width")),
        width_unit="s",
        height=height,
        height_unit="mAU",
        relative_height=try_float_or_none(peak.get(str, "PctHeight")),
        written_name=peak.get(str, "Name"),
        relative_peak_analyte_amount=try_float_or_none(peak.get(str, "PctAmount")),
        peak_analyte_amount=try_float_or_none(peak.get(str, "Amount")),
        index=peak.get(str, "PeakCounter"),
        baseline_value_at_start_of_peak=baseline_start,
        baseline_value_at_end_of_peak=baseline_end,
        relative_corrected_peak_area=try_float_or_none(peak.get(str, "CorrectedArea~")),
        custom_info=custom_info,
    )

    return peak_item


def _get_sample_role_type(sample_type: str | None) -> str | None:
    if not sample_type:
        return None

    sample_role_types = [
        srt.value for srt in SampleRoleType if str(sample_type).lower() in srt.value
    ]
    return (
        sample_role_types[0]
        if len(sample_role_types) > 0
        else SampleRoleType.unknown_sample_role.value
    )


def _build_data_processing_by_method_id(
    processing_methods: list[dict[str, Any]], processing_method_ids: set[str]
) -> defaultdict[str, list[dict[str, Any]]]:
    data_processing_by_processing_method_id: defaultdict[
        str, list[dict[str, Any]]
    ] = defaultdict(list)

    for proc_method in processing_methods:
        method_id = proc_method.get("id")

        if method_id and method_id in processing_method_ids:
            proc_json = JsonData(proc_method)
            field_mappings = {
                "name": (str, "name", None),
                "version": (str, "version", None),
                "locked": (str, "locked", None),
                "modified_by": (str, "modified_by", None),
                "revision_comment": (str, "revision comment", None),
                "revision_history": (str, "revision history", None),
                "component_nodes": (str, "component_nodes", None),
                "default_result_set": (str, "default_result_set", None),
                "average_by": (str, "average_by", None),
                "ccal_ref1": (str, "ccal_ref1", None),
                "comments": (str, "comments", None),
                "date": (str, "date", None),
                "include_int_std_amounts": (str, "include_int_std_amounts", None),
            }

            proc_method_info = proc_json.get_keys_as_dict(
                field_mappings,
                skip={"integration_parameters", "components"},
                include_unread=True,
            )

            # Add method_id directly since it doesn't come from get_keys_as_dict
            proc_method_info["id"] = str(method_id)

            integration_parameters = proc_method.get("integration_parameters")
            if integration_parameters:
                integration_param_fields = integration_parameters.get("fields")
                if integration_param_fields:
                    proc_method_info.update(integration_param_fields)

            components = proc_method.get("components")
            if components:
                _process_components(
                    components,
                    proc_method_info,
                    data_processing_by_processing_method_id,
                    method_id,
                )

    return data_processing_by_processing_method_id


def _process_components(
    components: list[dict[str, Any]],
    proc_method_info: dict[str, Any],
    data_processing_by_method_id: defaultdict[str, list[dict[str, Any]]],
    method_id: int,
) -> None:
    for component in components:
        fields = component.get("fields")
        if fields:
            fields.update(proc_method_info)
            data_processing_by_method_id[str(method_id)].append(filter_nulls(fields))


def _build_processing_items(
    results: list[dict[str, Any]],
    data_processing_by_method_id: defaultdict[str, list[dict[str, Any]]],
) -> list[ProcessedData]:
    processing_items = []

    for data in results:
        result_data = JsonData(data)
        field_mappings = {
            "integration_algorithm_type": (str, "IntegrationAlgorithm", None),
            "data_processing_method": (str, "ProcessingMethod", None),
            "data_processing_time": (str, "DateProcessed", None),
            "peak_width": (float, "PeakWidth", None),
            "retention_time": (float, "RetentionTime", None),
            "retention_time_window_width": (float, "RTWindow", None),
            "relative_response": (float, "RelativeResponse", None),
            "CalculationType": (str, "CalculationType", None),
            "CalibrationId": (str, "CalibrationId", None),
            "ProcessingLocked": (str, "ProcessingLocked", None),
            "Manual": (str, "Manual", None),
            "ProcessedBy": (str, "ProcessedBy", None),
            "ProcessedAs": (str, "ProcessedAs", None),
            "UseForPrecision": (str, "UseForPrecision", None),
            "PrepType": (str, "PrepType", None),
            "AnalysisMethod": (str, "AnalysisMethod", None),
            "IntegrationSystemPolicies": (str, "IntegrationSystemPolicies", None),
            "ProcessingMethodId": (str, "ProcessingMethodId", None),
            "ProcessedChannelType": (str, "ProcessedChannelType", None),
            "ProcessedChanDesc": (str, "ProcessedChanDesc", None),
            "PeakRatioReference": (str, "PeakRatioReference", None),
            "Threshold": (str, "Threshold", None),
            "SourceSoftwareInfo": (str, "SourceSoftwareInfo", None),
            "SampleValuesUsedinCalculations": (
                str,
                "SampleValuesUsedinCalculations",
                None,
            ),
            "NumOfResultsStored": (str, "NumOfResultsStored", None),
            "NumOfProcessOnlySampleSets": (str, "NumOfProcessOnlySampleSets", None),
            "Factor1": (str, "Factor1", None),
            "Factor2": (str, "Factor2", None),
            "Factor3": (str, "Factor3", None),
            "Factor1Operator": (str, "Factor1Operator", None),
            "Factor2Operator": (str, "Factor2Operator", None),
            "Factor3Operator": (str, "Factor3Operator", None),
            "ResultSetId": (str, "ResultSetId", None),
            "ResultSetName": (str, "ResultSetName", None),
            "ResultSetDate": (str, "ResultSetDate", None),
            "ResultId": (str, "ResultId", None),
            "ResultType": (str, "ResultType", None),
            "ResultComments": (str, "ResultComments", None),
            "ResultCodes": (str, "ResultCodes", None),
            "ResultNum": (str, "ResultNum", None),
            "ResultSampleSetMethod": (str, "ResultSampleSetMethod", None),
            "ResultSource": (str, "ResultSource", None),
            "ResultSuperseded": (str, "ResultSuperseded", None),
            "TotalArea": (str, "TotalArea", None),
            "TotalAdjArea": (str, "TotalAdjArea", None),
            "AdjustedTotalArea": (str, "AdjustedTotalArea", None),
            "TotalRS_AdjAreaPct": (str, "TotalRS_AdjAreaPct", None),
            "TotalRS_FinalResult": (str, "TotalRS_FinalResult", None),
            "Largest_NG_RS_AdjAreaPct": (str, "Largest_NG_RS_AdjAreaPct", None),
            "LargestNamedRS_FinalResult": (str, "LargestNamedRS_FinalResult", None),
            "LargestRS_FinalResult": (str, "LargestRS_FinalResult", None),
            "LargestUnnamedRS_FinalResult": (str, "LargestUnnamedRS_FinalResult", None),
            "Total_ICH_AdjArea": (str, "Total_ICH_AdjArea", None),
            "PercentUnknowns": (str, "PercentUnknowns", None),
            "NumberofImpurityPeaks": (str, "NumberofImpurityPeaks", None),
            "Faults": (str, "Faults", None),
        }

        processing_custom_info = result_data.get_keys_as_dict(
            field_mappings, include_unread=True
        )
        processing_method_id = result_data.get(str, "ProcessingMethodId")

        # Use an empty list as default when method_id is None or when it's not in the dictionary
        data_items = []
        if processing_method_id is not None:
            data_items = data_processing_by_method_id.get(processing_method_id, [])

        processing_items.append(
            ProcessedData(
                custom_info=processing_custom_info,
                data=data_items,
            )
        )

    return processing_items


def _create_measurements(
    injection: JsonData,
    metadata_fields: JsonData,
) -> list[Measurement]:
    peaks = injection.data.get("peaks", [])
    results = injection.data.get("results", [])

    sample_type = injection.get(str, "SampleType")
    sample_role_type = _get_sample_role_type(sample_type)
    measurement_time = metadata_fields.get(str, "SampleSetStartDate")

    # Get measurement custom info
    measurement_custom_info = injection.get_keys_as_dict(
        {
            "Comments": (str, "Comments", None),
            "DateAcquired": (str, "DateAcquired", None),
            "RunTime": (str, "RunTime", None),
        }
    )
    measurement_custom_info.update(
        metadata_fields.get_keys_as_dict(
            {
                "measurement_end_time": (str, "SampleSetFinishDate", None),
            }
        )
    )

    # Create device control docs
    device_control_field_mappings = {
        "SolventA": (str, "SolventA", None),
        "SolventB": (str, "SolventB", None),
        "SolventC": (str, "SolventC", None),
        "SolventD": (str, "SolventD", None),
        "ChannelDescription": (str, "ChannelDescription", None),
        "ChannelType": (str, "ChannelType", None),
        "ChannelStatus": (str, "ChannelStatus", None),
        "DataStart": (str, "DataStart", None),
        "DataEnd": (str, "DataEnd", None),
        "CirculationNo": (str, "CirculationNo", None),
        "device acquisition method": (str, "AcqMethodSet", None),
        "total measurement duration setting": (float, "RunTime", None),
        "Channel": (str, "Channel", None),
        "ScaletouV": (str, "ScaletouV", None),
        "SecondChannelId": (str, "SecondChannelId", None),
    }

    device_control_custom_info = injection.get_keys_as_dict(
        device_control_field_mappings
    )

    instrument_methods = metadata_fields.data.get("instrument_methods", [])
    if instrument_methods:
        instrument_method_id = injection.get(str, "InstrumentMethodId")
        if instrument_method_id:
            for method in instrument_methods:
                method_id = method.get("id")
                if method_id and str(method_id) == str(instrument_method_id):
                    instrument_method_field_mapping = {
                        "device method identifier": (str, "id", None),
                        "name": (str, "name", None),
                        "comments": (str, "comments", None),
                        "methods_date": (str, "date", None),
                        "InstOnStatus": (str, "InstOnStatus", None),
                        "type": (str, "type", None),
                    }
                    temp_json_data = JsonData(method)
                    device_control_custom_info.update(
                        temp_json_data.get_keys_as_dict(
                            instrument_method_field_mapping, include_unread=True
                        )
                    )

    device_control_docs = [
        DeviceControlDoc(
            device_type=constants.DEVICE_TYPE,
            device_identifier=injection.get(str, "ChannelId"),
            detector_sampling_rate_setting=injection.get(float, "SamplingRate"),
            device_control_custom_info=device_control_custom_info,
        )
    ]

    processing_method_ids = set()
    for result_item in results:
        if method_id := result_item.get("ProcessingMethodId"):
            processing_method_ids.add(method_id)

    processing_methods = metadata_fields.data.get("processing_methods", [])
    data_processing_by_method_id = (
        _build_data_processing_by_method_id(processing_methods, processing_method_ids)
        if processing_methods and processing_method_ids
        else defaultdict(list)
    )

    processing_items = _build_processing_items(results, data_processing_by_method_id)

    # Get sample and injection custom info
    sample_custom_info = injection.get_keys_as_dict(
        {
            "SampleConcentration": (str, "SampleConcentration", None),
            "ELN_SampleID": (str, "ELN_SampleID", None),
            "PrepType": (str, "PrepType", None),
            "SampleSetAcquiring": (str, "SampleSetAcquiring", None),
            "SampleSetAltered": (str, "SampleSetAltered", None),
            "SampleSetCurrentId": (str, "SampleSetCurrentId", None),
            "OriginalSampleSetId": (str, "OriginalSampleSetId", None),
            "OriginalVialId": (str, "OriginalVialId", None),
            "Vial": (str, "Vial", None),
            "lot_number": (str, "Lot", None),
            "sample_weight": (float, "SampleWeight", None),
            "Label": (str, "Label", None),
            "VialId": (str, "VialId", None),
            "VialIdResult": (str, "VialIdResult", None),
            "SampleType": (str, "SampleType", None),
            "dilution_factor_setting": (float, "Dilution", None),
        }
    )

    injection_custom_info = injection.get_keys_as_dict(
        {
            "InjectionStatus": (str, "InjectionStatus", None),
            "InjectionType": (str, "InjectionType", None),
            "Injection": (str, "Injection", None),
            "Volume": (str, "Volume", None),
        }
    )

    # Get required values
    sample_name = injection[str, "SampleName"]
    date_acquired = injection[str, "DateAcquired"]

    injection.mark_read({"chrom", "peaks", "results", "curves"})

    # Create and return the measurement
    return [
        Measurement(
            measurement_identifier=random_uuid_str(),
            sample_identifier=sample_name,
            batch_identifier=metadata_fields.get(str, "sample_set_id"),
            sample_role_type=sample_role_type,
            written_name=injection.get(str, "Label"),
            chromatography_serial_num=injection.get(
                str, "ColumnSerialNumber", NOT_APPLICABLE
            ),
            autosampler_injection_volume_setting=injection.get(
                float, "InjectionVolume"
            ),
            injection_identifier=injection[str, "InjectionId"],
            injection_time=date_acquired,
            peaks=[
                peak
                for peak in [_create_peak(JsonData(peak_dict)) for peak_dict in peaks]
                if peak
            ],
            chromatogram_data_cube=_get_chromatogram(injection),
            device_control_docs=device_control_docs,
            measurement_time=measurement_time,
            location_identifier=injection.get(str, "VialId"),
            flow_rate=try_float_or_none(injection.get(str, "FlowRate")),
            measurement_custom_info={
                **measurement_custom_info,
                **injection.get_unread(),
            },
            processed_data=processing_items,
            sample_custom_info=sample_custom_info,
            injection_custom_info=injection_custom_info,
        )
    ]


def _group_injections_by_sample(
    injections: list[JsonData],
) -> defaultdict[str, list[JsonData]]:
    sample_to_injection: defaultdict[str, list[JsonData]] = defaultdict(list)

    for injection in injections:
        sample_name = injection[str, "SampleName"]
        sample_to_injection[sample_name].append(injection)

    return sample_to_injection


def create_measurement_groups(
    injections: list[JsonData],
    metadata_fields: JsonData,
) -> list[MeasurementGroup]:
    sample_to_injection = _group_injections_by_sample(injections)

    measurement_aggregate_field_mappings = {
        "Node": (str, "Node", None),
        "measurement_method_identifier": (str, "SampleSetMethod", None),
        "SampleSetName": (str, "SampleSetName", None),
        "SampleSetType": (str, "SampleSetType", None),
        "SampleSetComments": (str, "SampleSetComments", None),
    }

    measurement_aggregate_custom_info = metadata_fields.get_keys_as_dict(
        measurement_aggregate_field_mappings
    )

    measurement_groups = []
    for _, sample_injections in sample_to_injection.items():
        if not sample_injections:
            continue

        first_injection = sample_injections[0]

        group_field_mappings = {
            "Altered": (str, "Altered", None),
            "ELN_DocumentID": (str, "ELN_DocumentID", None),
            "ELN_SectionGUID": (str, "ELN_SectionGUID", None),
            "InstrumentMethodName": (str, "InstrumentMethodName", None),
            "InstrumentMethodId": (str, "InstrumentMethodId", None),
            "SSM_Pattern": (str, "SSM_Pattern", None),
            "Superseded": (str, "Superseded", None),
            "SummaryFaults": (str, "SummaryFaults", None),
        }

        group_custom_info = first_injection.get_keys_as_dict(group_field_mappings)
        group_custom_info.update(measurement_aggregate_custom_info)

        measurements = [
            measurement
            for injection in sample_injections
            for measurement in _create_measurements(injection, metadata_fields)
        ]

        measurement_groups.append(
            MeasurementGroup(
                measurements=measurements,
                measurement_aggregate_custom_info=group_custom_info,
            )
        )

    return measurement_groups



from dataclasses import dataclass

from allotropy.allotrope.converter import add_custom_information_document
from allotropy.allotrope.models.adm.mass_spectrometry.rec._2025._06.mass_spectrometry import (
    DataProcessingAggregateDocument,
    DataProcessingDocumentItem,
    DetectorControlAggregateDocument,
    DetectorControlDocumentItem,
    DeviceDocumentItem,
    DeviceSystemDocument,
    MassSpectrometryAggregateDocument,
    MassSpectrometryDocumentItem,
    MeasurementDocument,
    Model,
    PeakItem,
    PeakList,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    SampleIntroductionDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDalton,
    TQuantityValueHertz,
    TQuantityValueMassPerCharge,
    TQuantityValueMilliliterPerMinute,
    TQuantityValuePercent,
    TQuantityValueSecondTime,
)
from allotropy.allotrope.models.shared.definitions.definitions import TQuantityValue
from allotropy.allotrope.schema_mappers.schema_mapper import SchemaMapper
from allotropy.parsers.utils.values import quantity_or_none


@dataclass(frozen=True)
class Device:
    device_type: str
    model_number: str
    product_manufacturer: str
    device_custom_info: dict[str, object] | None = None


@dataclass(frozen=True)
class DetectorControl:
    detection_type: str
    detection_duration_setting: float | None = None
    detector_relative_offset_setting: float | None = None
    detector_sampling_rate_setting: float | None = None
    m_z_maximum_setting: float | None = None
    m_z_minimum_setting: float | None = None
    polarity_setting: str | None = None


@dataclass(frozen=True)
class SampleIntroduction:
    sample_introduction_medium: str
    sample_introduction_mode_setting: str
    flow_rate_setting: float | None = None
    laser_firing_frequency_setting: float | None = None
    sample_introduction_description: str | None = None


@dataclass(frozen=True)
class Measurement:
    analyst: str
    submitter: str
    measurement_mode_setting: str

    # Optional associations
    sample_identifier: str | None = None
    detector_control: DetectorControl | None = None
    sample_introduction: SampleIntroduction | None = None

    # Optional aggregates
    processed_data_types: list[str] | None = None
    data_processing_types: list[str] | None = None


@dataclass(frozen=True)
class Peak:
    identifier: str | None = None
    m_z: float | None = None
    mass: float | None = None
    peak_area_value: float | None = None
    peak_area_unit: str | None = None
    peak_height_value: float | None = None
    peak_height_unit: str | None = None
    peak_width_value: float | None = None
    peak_width_unit: str | None = None
    relative_peak_area: float | None = None
    relative_peak_height: float | None = None
    written_name: str | None = None


@dataclass(frozen=True)
class Metadata:
    asset_management_identifier: str
    data_processing_description: str | None = None
    devices: list[Device] | None = None
    device_system_custom_info: dict[str, object] | None = None


@dataclass(frozen=True)
class Data:
    metadata: Metadata
    measurements: list[Measurement]
    peaks: list[Peak] | None = None


class Mapper(SchemaMapper[Data, Model]):
    MANIFEST = "http://purl.allotrope.org/manifests/mass-spectrometry/REC/2025/06/mass-spectrometry.manifest"

    def map_model(self, data: Data) -> Model:
        return Model(
            manifest=self.MANIFEST,
            data_processing_description=data.metadata.data_processing_description,
            mass_spectrometry_aggregate_document=MassSpectrometryAggregateDocument(
                device_system_document=self._get_device_system_document(data.metadata),
                mass_spectrometry_document=[
                    self._get_mass_spectrometry_document_item(m)
                    for m in data.measurements
                ],
            ),
            peak_list=self._get_peak_list(data.peaks),
        )

    def _get_device_system_document(self, metadata: Metadata) -> DeviceSystemDocument:
        return add_custom_information_document(
            DeviceSystemDocument(
                asset_management_identifier=metadata.asset_management_identifier,
                device_document=[
                    DeviceDocumentItem(
                        device_type=device.device_type,
                        model_number=device.model_number,
                        product_manufacturer=device.product_manufacturer,
                    )
                    for device in (metadata.devices or [])
                ]
                or None,
            ),
            metadata.device_system_custom_info,
        )

    def _get_mass_spectrometry_document_item(
        self, measurement: Measurement
    ) -> MassSpectrometryDocumentItem:
        return MassSpectrometryDocumentItem(
            analyst=measurement.analyst,
            submitter=measurement.submitter,
            sample_document=(
                SampleDocument(sample_identifier=measurement.sample_identifier)
                if measurement.sample_identifier
                else None
            ),
            sample_introduction_document=self._get_sample_introduction_document(
                measurement.sample_introduction
            ),
            measurement_document=self._get_measurement_document(measurement),
            processed_data_aggregate_document=self._get_processed_data_aggregate_document(
                measurement.processed_data_types
            ),
            data_processing_aggregate_document=self._get_data_processing_aggregate_document(
                measurement.data_processing_types
            ),
        )

    def _get_sample_introduction_document(
        self, intro: SampleIntroduction | None
    ) -> SampleIntroductionDocument | None:
        if not intro:
            return None

        return SampleIntroductionDocument(
            sample_introduction_medium=intro.sample_introduction_medium,
            sample_introduction_mode_setting=intro.sample_introduction_mode_setting,
            flow_rate_setting=quantity_or_none(
                TQuantityValueMilliliterPerMinute, intro.flow_rate_setting
            ),
            laser_firing_frequency_setting=quantity_or_none(
                TQuantityValueHertz, intro.laser_firing_frequency_setting
            ),
            sample_introduction_description=intro.sample_introduction_description,
        )

    def _get_measurement_document(
        self, measurement: Measurement
    ) -> MeasurementDocument:
        return MeasurementDocument(
            measurement_mode_setting=measurement.measurement_mode_setting,
            detector_control_aggregate_document=self._get_detector_control_aggregate_document(
                measurement.detector_control
            ),
        )

    def _get_detector_control_aggregate_document(
        self, control: DetectorControl | None
    ) -> DetectorControlAggregateDocument | None:
        if not control:
            return None
        return DetectorControlAggregateDocument(
            detector_control_document=[
                DetectorControlDocumentItem(
                    detection_type=control.detection_type,
                    detection_duration_setting=quantity_or_none(
                        TQuantityValueSecondTime, control.detection_duration_setting
                    ),
                    detector_relative_offset_setting=quantity_or_none(
                        TQuantityValueSecondTime,
                        control.detector_relative_offset_setting,
                    ),
                    detector_sampling_rate_setting=quantity_or_none(
                        TQuantityValueHertz, control.detector_sampling_rate_setting
                    ),
                    m_z_maximum_setting=quantity_or_none(
                        TQuantityValueMassPerCharge, control.m_z_maximum_setting
                    ),
                    m_z_minimum_setting=quantity_or_none(
                        TQuantityValueMassPerCharge, control.m_z_minimum_setting
                    ),
                    polarity_setting=control.polarity_setting,
                )
            ]
        )

    def _get_processed_data_aggregate_document(
        self, types: list[str] | None
    ) -> ProcessedDataAggregateDocument | None:
        if not types:
            return None
        return ProcessedDataAggregateDocument(
            processed_data_document=[
                ProcessedDataDocumentItem(data_format_specification_type=t)
                for t in types
            ]
        )

    def _get_data_processing_aggregate_document(
        self, types: list[str] | None
    ) -> DataProcessingAggregateDocument | None:
        if not types:
            return None
        return DataProcessingAggregateDocument(
            data_processing_document=[
                DataProcessingDocumentItem(data_processing_type=t) for t in types
            ]
        )

    def _get_peak_list(self, peaks: list[Peak] | None) -> PeakList | None:
        if not peaks:
            return None
        return PeakList(
            peak=[self._get_peak_item(p) for p in peaks] or None,
        )

    def _get_peak_item(self, peak: Peak) -> PeakItem:
        return PeakItem(
            identifier=peak.identifier,
            m_z=quantity_or_none(TQuantityValueMassPerCharge, peak.m_z),
            mass=quantity_or_none(TQuantityValueDalton, peak.mass),
            peak_area=(
                TQuantityValue(value=peak.peak_area_value, unit=peak.peak_area_unit)
                if peak.peak_area_value is not None and peak.peak_area_unit is not None
                else None
            ),
            peak_height=(
                TQuantityValue(value=peak.peak_height_value, unit=peak.peak_height_unit)
                if peak.peak_height_value is not None
                and peak.peak_height_unit is not None
                else None
            ),
            peak_width=(
                TQuantityValue(value=peak.peak_width_value, unit=peak.peak_width_unit)
                if peak.peak_width_value is not None
                and peak.peak_width_unit is not None
                else None
            ),
            relative_peak_area=quantity_or_none(
                TQuantityValuePercent, peak.relative_peak_area
            ),
            relative_peak_height=quantity_or_none(
                TQuantityValuePercent, peak.relative_peak_height
            ),
            written_name=peak.written_name,
        )



from typing import IO

import chardet

from allotropy.constants import CHARDET_ENCODING, DEFAULT_ENCODING
from allotropy.exceptions import AllotropeConversionError, AllotropeParsingError


def _get_contents(contents: bytes | IO[bytes] | IO[str]) -> str | bytes:
    if isinstance(contents, bytes):
        return contents
    actual_contents = contents.read()
    contents.seek(0)
    return actual_contents


def determine_encoding(
    contents: bytes | IO[bytes] | IO[str], encoding: str | None
) -> list[str | None]:
    if not encoding:
        return [DEFAULT_ENCODING]
    if encoding != CHARDET_ENCODING:
        return [encoding]

    actual_contents = _get_contents(contents)
    if isinstance(actual_contents, str):
        return [None]
    if not actual_contents:
        msg = "Unable to detect encoding for empty bytes string, file may be empty."
        raise AllotropeConversionError(msg)

    detect_result = chardet.detect(actual_contents)
    if not detect_result["encoding"]:
        msg = (
            f"Unable to detect text encoding for file with content: {actual_contents!r}"
        )
        raise AllotropeParsingError(msg)

    detected_encoding = detect_result["encoding"]
    confidence = detect_result["confidence"]

    # chardet can misdetect UTF-8 multi-byte sequences as ISO-8859-1 or similar
    # Latin-1 encodings, especially in older versions (5.x). Always try UTF-8
    # first when Latin-1 family encodings are detected to avoid mojibake.
    latin1_encodings = {
        "ISO-8859-1",
        "ISO-8859-2",
        "ISO-8859-3",
        "ISO-8859-4",
        "ISO-8859-5",
        "ISO-8859-6",
        "ISO-8859-7",
        "ISO-8859-8",
        "ISO-8859-9",
        "ISO-8859-10",
        "ISO-8859-13",
        "ISO-8859-14",
        "ISO-8859-15",
        "ISO-8859-16",
        "cp1250",
        "cp1251",
        "cp1252",
        "cp1253",
        "cp1254",
        "cp1255",
        "cp1256",
        "cp1257",
        "cp1258",
        "windows-1250",
        "windows-1251",
        "windows-1252",
        "windows-1253",
        "windows-1254",
        "windows-1255",
        "windows-1256",
        "windows-1257",
        "windows-1258",
    }

    # Normalize encoding name for comparison
    normalized_encoding = detected_encoding.upper() if detected_encoding else ""

    if confidence < 0.3:
        # Very low confidence, try multiple encodings
        encodings: list[str | None] = [DEFAULT_ENCODING, "windows-1252"]
        if detected_encoding and detected_encoding not in encodings:
            encodings.append(detected_encoding)
        return encodings
    elif normalized_encoding in latin1_encodings or confidence < 0.7:
        # For Latin-1 family or medium confidence, try UTF-8 first
        # This handles chardet 5.x incorrectly detecting UTF-8 as ISO-8859-1
        if detected_encoding != DEFAULT_ENCODING:
            return [DEFAULT_ENCODING, detected_encoding]
        return [DEFAULT_ENCODING]
    else:
        # High confidence non-Latin encoding, use as detected
        return [detected_encoding]


def decode(contents: IO[bytes] | IO[str], encoding: str | None) -> str:
    actual_contents = _get_contents(contents)
    if isinstance(actual_contents, str):
        return actual_contents
    possible_encodings = determine_encoding(actual_contents, encoding)

    for encoding in possible_encodings:
        # NOTE: this should not be possible, we only return None if contents is str, which should have returned already.
        if encoding is None:
            msg = f"Could not determine encoding of contents: {actual_contents!r}"
            raise AssertionError(msg)
        try:
            decoded = actual_contents.decode(encoding)
            # Strip BOM (Byte Order Mark) if present
            # UTF-16 and UTF-8 files may include BOM character U+FEFF
            # which should be removed from the content
            if decoded and decoded[0] == "\ufeff":
                decoded = decoded[1:]
            return decoded
        except UnicodeDecodeError as e:
            if encoding != possible_encodings[-1]:
                continue
            msg = f"Unable to decode bytes with encoding '{encoding}' with error: {e}, bytes: {actual_contents!r}"
            raise AllotropeParsingError(msg) from e
        except LookupError as e:
            msg = f"Invalid encoding: '{encoding}'."
            raise AllotropeConversionError(msg) from e
    msg = f"Unable to decode contents with possible encodings: {possible_encodings}. Contents: {actual_contents!r}"
    raise AllotropeConversionError(msg)



from dataclasses import dataclass

from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    Measurement,
)


@dataclass
class CalculatedDataStructure:
    measurement: Measurement
    average_of_blank_used: float | None
    corrected_value: float | None



from allotropy.allotrope.models.adm.multi_analyte_profiling.benchling._2024._09.multi_analyte_profiling import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.multi_analyte_profiling.benchling._2024._09.multi_analyte_profiling import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.luminex_intelliflex.constants import DISPLAY_NAME
from allotropy.parsers.luminex_intelliflex.luminex_intelliflex_structure import (
    create_metadata,
)
from allotropy.parsers.luminex_xponent.luminex_xponent_reader import (
    LuminexXponentReader,
)
from allotropy.parsers.luminex_xponent.luminex_xponent_structure import (
    create_measurement_groups,
    Data as XponentData,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class LuminexIntelliflexParser(VendorParser[Data, Model]):
    DISPLAY_NAME = DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = LuminexXponentReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = LuminexXponentReader(named_file_contents)
        data = XponentData.create(reader)
        return Data(
            create_metadata(
                data.header, data.calibrations, named_file_contents.original_file_path
            ),
            *create_measurement_groups(data.measurement_list.measurements, data.header),
        )



from allotropy.calcdocs.extractor import Element, Extractor
from allotropy.parsers.cytiva_biacore_t200_control.cytiva_biacore_t200_control_structure import (
    ReportPointData,
    SampleData,
)


class CytivaBiacoreExtractor(Extractor[ReportPointData]):
    @classmethod
    def sample_data_to_elements(cls, sample_data: SampleData) -> list[Element]:
        elements = []
        for measurements_data in sample_data.measurements.values():
            for measurement in measurements_data:
                if measurement.report_point_data is None:
                    continue
                elements.extend(cls.get_elements(measurement.report_point_data))
        return elements

    @classmethod
    def get_elements(cls, report_points: list[ReportPointData]) -> list[Element]:
        return [cls.to_element(report_point) for report_point in report_points]

    @classmethod
    def to_element(cls, report_point: ReportPointData) -> Element:
        return Element(
            uuid=report_point.identifier,
            data={
                "absolute_resonance": report_point.absolute_resonance,
                "min_resonance": report_point.min_resonance,
                "max_resonance": report_point.max_resonance,
                "lrsd": report_point.lrsd,
                "slope": report_point.slope,
                "sd": report_point.sd,
            },
        )



from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2025._03.plate_reader import (
    Measurement,
)
from allotropy.calcdocs.extractor import Element, Extractor


class LunaticExtractor(Extractor[Measurement]):
    @classmethod
    def to_element(cls, measurement: Measurement) -> Element:
        custom_info = (
            measurement.calc_docs_custom_info
            if measurement.calc_docs_custom_info
            else {}
        )

        return Element(
            uuid=measurement.identifier,
            data={
                "uuid": measurement.identifier,
                "wavelength id": "NA",
                "absorbance": 0.0,
                "detection type": measurement.detection_type,
                **custom_info,
            },
        )



from __future__ import annotations

from dataclasses import dataclass
import itertools
from pathlib import Path
import re
from typing import Any

from dateutil.parser import parse, ParserError
import pandas as pd

from allotropy.allotrope.schema_mappers.adm.multi_analyte_profiling.benchling._2024._09.multi_analyte_profiling import (
    Analyte,
    Calibration,
    Error,
    Measurement as MapperMeasurement,
    MeasurementGroup,
    Metadata,
    StatisticDimension,
    StatisticsDocument,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.constants import NEGATIVE_ZERO
from allotropy.parsers.luminex_xponent.constants import (
    CALCULATED_DATA_SECTIONS,
    DEFAULT_CONTAINER_TYPE,
    DEFAULT_DEVICE_TYPE,
    DEFAULT_SOFTWARE_NAME,
    EXPECTED_CALIBRATION_RESULT_LEN,
    MINIMUM_CALIBRATION_LINE_COLS,
    REQUIRED_SECTIONS,
    STATISTIC_SECTIONS_CONF,
)
from allotropy.parsers.luminex_xponent.luminex_xponent_reader import (
    LuminexXponentReader,
)
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
    DataSource,
    Referenceable,
)
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float,
    try_non_nan_float_or_negative_zero,
    try_non_nan_float_or_none,
)


@dataclass(frozen=True)
class Header:
    model_number: str | None
    software_version: str | None
    equipment_serial_number: str | None
    analytical_method_identifier: str | None
    method_version: str | None
    experimental_data_identifier: str | None
    sample_volume_setting: float | None
    plate_well_count: float
    measurement_time: str
    data_system_instance_identifier: str
    detector_gain_setting: str | None
    minimum_assay_bead_count_setting: float | None
    analyst: str | None
    custom_info: dict[str, Any]

    @classmethod
    def create(
        cls,
        header_data: pd.DataFrame,
        header_row: SeriesData,
        minimum_assay_bead_count_setting: float | None,
    ) -> Header:
        sample_volume = header_row.get(str, ["SampleVolume", "MaxSampleUptakeVolume"])

        header_row.mark_read({"Program", "ProtocolPlate", "Date"})

        return Header(
            model_number=cls._get_model_number(header_data),
            software_version=header_row.get(str, "Build"),
            equipment_serial_number=header_row.get(str, "SN"),
            analytical_method_identifier=header_row.get(str, "ProtocolName"),
            method_version=header_row.get(str, "ProtocolVersion"),
            experimental_data_identifier=header_row.get(str, "Batch"),
            sample_volume_setting=(
                try_float(sample_volume.split()[0], "sample volume setting")
                if sample_volume
                else None
            ),
            plate_well_count=cls._get_plate_well_count(header_data),
            measurement_time=header_row[str, "BatchStartTime"],
            detector_gain_setting=header_row.get(
                str, ["ProtocolReporterGain", "ProtocolOperatingMode"]
            ),
            data_system_instance_identifier=header_row[str, "ComputerName"],
            minimum_assay_bead_count_setting=minimum_assay_bead_count_setting,
            analyst=header_row.get(str, "Operator"),
            custom_info={
                "Country Code": header_row.get(str, "Country Code"),
                "ProtocolDevelopingCompany": header_row.get(
                    str, "ProtocolDevelopingCompany"
                ),
                "Version": header_row.get(str, "Version"),
                # Used for the measurement groups
                "BatchStopTime": header_row.get(str, "BatchStopTime"),
                "ProtocolDescription": header_row.get(str, "ProtocolDescription"),
            },
        )

    @classmethod
    def _get_model_number(cls, header_data: pd.DataFrame) -> str | None:
        if "Program" not in header_data:
            return None
        program_data = header_data["Program"]

        try:
            model_number = program_data.iloc[2]
        except IndexError as e:
            msg = (
                "Unable to find model number in Program row, expected value at index 2"
            )
            raise AllotropeConversionError(msg) from e

        return str(model_number) if model_number else None

    @classmethod
    def _get_plate_well_count(cls, header_data: pd.DataFrame) -> float:
        if "ProtocolPlate" not in header_data:
            msg = "Unable to find required value 'ProtocolPlate' data in header block."
            raise AllotropeConversionError(msg)
        protocol_plate_data = header_data["ProtocolPlate"]

        try:
            plate_well_count = protocol_plate_data.iloc[3]
        except IndexError as e:
            msg = "Unable to find plate well count in ProtocolPlate row, expected value at index 3."
            raise AllotropeConversionError(msg) from e

        return try_float(str(plate_well_count), "plate well count")


def create_calibration(calibration_data: SeriesData) -> Calibration:
    """Create a CalibrationItem from a calibration line.

    Each line should follow one of these patterns:
    - "Last <calibration_name>,<calibration_report> <calibration_time><,,,,"
    - "Last <calibration_name>,<calibration_time><,,,,"
    """
    if len(calibration_data.series.index) < MINIMUM_CALIBRATION_LINE_COLS:
        msg = f"Expected at least two columns on the calibration line, got:\n{calibration_data.series}."
        raise AllotropeConversionError(msg)

    # Read the calibration data using SeriesData methods
    calibration_name = calibration_data.series.iloc[0]
    calibration_info = str(calibration_data.series.iloc[1]).strip()
    calibration_data.get_unread()

    # Check if the calibration info starts with a known status value
    status_values = ["Passed", "Failed", "Calibrated", "Verified"]
    first_word = calibration_info.split()[0] if calibration_info.split() else ""

    report = None
    time = None

    if first_word in status_values:
        calibration_result = calibration_info.split(maxsplit=1)
        if len(calibration_result) == EXPECTED_CALIBRATION_RESULT_LEN:
            report = calibration_result[0]
            time = calibration_result[1]
    else:
        try:
            parse(calibration_info)
            time = calibration_info
        except ParserError:
            pass

    if not time:
        msg = f"Invalid calibration result format, got: ['{calibration_info}']."
        raise AllotropeConversionError(msg)

    return Calibration(
        name=calibration_name.replace("Last", "").strip(),
        time=time,
        report=report,
    )


@dataclass(frozen=True)
class Measurement:
    identifier: str
    sample_identifier: str
    location_identifier: str
    dilution_factor_setting: float
    assay_bead_count: float
    analytes: list[Analyte]
    calculated_data: list[CalculatedDocument]
    errors: list[Error] | None = None
    measurement_custom_info: dict[str, Any] | None = None
    sample_custom_info: dict[str, Any] | None = None
    device_control_custom_info: dict[str, Any] | None = None

    @classmethod
    def create(
        cls,
        results_data: dict[str, pd.DataFrame],
        count_data: SeriesData,
        bead_ids_data: SeriesData,
        header_row: SeriesData,
    ) -> Measurement:
        location = str(count_data.series.name)
        dilution_factor_data = results_data["Dilution Factor"]
        errors_data = results_data.get("Warnings/Errors")
        measurement_identifier = random_uuid_str()

        if location not in dilution_factor_data.index:
            msg = f"Could not find 'Dilution Factor' data for: '{location}'."
            raise AllotropeConversionError(msg)

        # Keys in the median data that are not analyte data.
        metadata_keys = ["Sample", "Total Events"]

        well_location, location_id = cls._get_location_details(location)
        dilution_factor_series = SeriesData(dilution_factor_data.loc[location])
        dilution_factor_setting = dilution_factor_series.get(
            float, "Dilution Factor", NEGATIVE_ZERO, validate=SeriesData.NOT_NAN
        )
        # Capture unread dilution factor data to prevent warnings
        dilution_factor_series.get_unread()
        errors: list[Error] = []
        data_errors = cls._get_errors(errors_data, well_location) or []
        for data_error in data_errors:
            errors.append(Error(error=data_error))

        if dilution_factor_setting == NEGATIVE_ZERO:
            errors.append(
                Error(error="Not reported in file", feature="dilution factor setting")
            )

        def get_statistics_error(
            raw_value: Any, analyte: str, role: str
        ) -> Error | None:
            """Check if the raw value is empty or NaN and return an Error if so."""
            error = None
            if raw_value == "":
                error = "Not reported"
            elif try_non_nan_float_or_none(raw_value) is None:
                error = "NaN"
            return Error(error=error, feature=f"{analyte} <{role}>") if error else None

        def get_statistic_dimensions(analyte: str) -> list[StatisticDimension]:
            statistic_dimensions = []
            for section, statistic_conf in STATISTIC_SECTIONS_CONF.items():
                if (statistic_table := results_data.get(section)) is None:
                    continue
                raw_value = statistic_table.at[location, analyte]
                role = statistic_conf.role
                if error := get_statistics_error(raw_value, analyte, role.value):
                    errors.append(error)
                statistic_dimensions.append(
                    StatisticDimension(
                        value=try_non_nan_float_or_negative_zero(raw_value),
                        unit=statistic_conf.unit,
                        statistic_datum_role=role,
                    )
                )
            return statistic_dimensions

        analytes = []
        calculated_data = []
        analyte_keys = [
            key for key in count_data.series.index if key not in metadata_keys
        ]
        for analyte in analyte_keys:
            assay_bead_identifier = bead_ids_data.get(str, analyte, "N/A")
            analytes.append(
                Analyte(
                    identifier=(analyte_identifier := random_uuid_str()),
                    name=analyte,
                    assay_bead_identifier=assay_bead_identifier,
                    assay_bead_count=count_data[float, analyte],
                    statistics=[
                        StatisticsDocument(
                            statistical_feature="fluorescence",
                            statistic_dimensions=get_statistic_dimensions(analyte),
                        ),
                    ],
                )
            )

            for section_name, unit in CALCULATED_DATA_SECTIONS.items():
                calculated_data_section = results_data.get(section_name, pd.DataFrame())
                # Some sections don't include the location as an index, in those cases we cannot associate
                # the calculated data with the measurement.
                if location not in calculated_data_section.index:
                    continue
                raw_value = calculated_data_section.at[location, analyte]
                if error := get_statistics_error(raw_value, analyte, section_name):
                    errors.append(error)
                calculated_data_id = random_uuid_str()
                calculated_data.append(
                    CalculatedDocument(
                        uuid=calculated_data_id,
                        name=section_name,
                        value=try_non_nan_float_or_negative_zero(raw_value),
                        unit=unit,
                        data_sources=[
                            DataSource(
                                feature="fluorescence",
                                reference=Referenceable(uuid=analyte_identifier),
                            )
                        ],
                    )
                )

        device_control_custom_info = {
            "ProtocolHeater": header_row.get(str, "ProtocolHeater"),
            "DDGate": header_row.get(str, "DDGate"),
            "SampleTimeout": header_row.get(str, "SampleTimeout"),
            "ProtocolAnalysis": header_row.get(str, "ProtocolAnalysis"),
            "ProtocolMicrosphere": header_row.get(str, "ProtocolMicrosphere"),
            "PlateReadDirection": header_row.get(str, "PlateReadDirection"),
        }
        sample_custom_info = {
            "BatchDescription": header_row.get(str, "BatchDescription"),
            "PanelName": header_row.get(str, "PanelName"),
            "BeadType": header_row.get(str, "BeadType"),
        }

        return Measurement(
            identifier=measurement_identifier,
            sample_identifier=count_data[str, "Sample"],
            location_identifier=location_id,
            dilution_factor_setting=dilution_factor_setting,
            assay_bead_count=count_data[float, "Total Events"],
            analytes=analytes,
            errors=errors,
            calculated_data=calculated_data,
            device_control_custom_info=device_control_custom_info,
            sample_custom_info=sample_custom_info,
            measurement_custom_info=count_data.get_unread() | header_row.get_unread(),
        )

    @classmethod
    def _get_location_details(cls, location: str) -> tuple[str, str]:
        location_regex = r"\d+\((?P<well_location>\d+,(?P<location_id>\w+))\)"
        match = assert_not_none(
            re.search(location_regex, location),
            msg=f"Invalid location format: {location}",
        )

        return match.group("well_location"), match.group("location_id")

    @classmethod
    def _get_errors(
        cls, errors_data: pd.DataFrame | None, well_location: str
    ) -> list[str] | None:
        if errors_data is None or well_location not in errors_data.index:
            return None

        def extract_message(data: SeriesData) -> str:
            message = data[str, "Message"]
            # Capture unread error data to prevent warnings
            data.get_unread()
            return message

        return map_rows(errors_data.loc[[well_location]], extract_message)


@dataclass(frozen=True)
class MeasurementList:
    measurements: list[Measurement]

    @classmethod
    def create(
        cls, results_data: dict[str, pd.DataFrame], header_row: SeriesData
    ) -> MeasurementList:
        if missing_sections := [
            section for section in REQUIRED_SECTIONS if section not in results_data
        ]:
            msg = f"Unable to parse input file, missing expected sections: {missing_sections}."
            raise AllotropeConversionError(msg)

        # Validate that there's at least one statistic section in results_data
        if not any(
            section in results_data for section in STATISTIC_SECTIONS_CONF.keys()
        ):
            msg = f"Unable to parse input file, expecting at least one of the following sections: {STATISTIC_SECTIONS_CONF.keys()}."
            raise AllotropeConversionError(msg)

        if "BeadID:" not in results_data["Units"].index:
            msg = (
                "Could not parse bead data from 'Units' section, missing 'BeadID:' row"
            )
            raise AllotropeConversionError()
        bead_ids_data = SeriesData(results_data["Units"].loc["BeadID:"])

        def create_measurement(count_data: SeriesData) -> Measurement:
            return Measurement.create(
                results_data=results_data,
                count_data=count_data,
                bead_ids_data=bead_ids_data,
                header_row=header_row,
            )

        return MeasurementList(map_rows(results_data["Count"], create_measurement))


@dataclass(frozen=True)
class Data:
    header: Header
    calibrations: list[Calibration]
    measurement_list: MeasurementList

    @classmethod
    def create(cls, reader: LuminexXponentReader) -> Data:
        header_row = SeriesData(reader.header_data.iloc[0])

        return Data(
            header=Header.create(
                reader.header_data, header_row, reader.minimum_assay_bead_count_setting
            ),
            calibrations=map_rows(reader.calibration_data, create_calibration),
            measurement_list=MeasurementList.create(reader.results_data, header_row),
        )


def create_metadata(
    header: Header, calibrations: list[Calibration], file_path: str
) -> Metadata:
    path = Path(file_path)
    return Metadata(
        file_name=path.name,
        asm_file_identifier=path.with_suffix(".json").name,
        unc_path=file_path,
        equipment_serial_number=header.equipment_serial_number,
        model_number=header.model_number,
        calibrations=calibrations,
        data_system_instance_identifier=header.data_system_instance_identifier,
        software_name=DEFAULT_SOFTWARE_NAME,
        software_version=header.software_version,
        device_type=DEFAULT_DEVICE_TYPE,
        custom_info=header.custom_info,
    )


def create_measurement_groups(
    measurements: list[Measurement], header: Header
) -> tuple[list[MeasurementGroup], list[CalculatedDocument] | None]:
    measurement_groups = [
        MeasurementGroup(
            analyst=header.analyst,
            analytical_method_identifier=header.analytical_method_identifier,
            method_version=header.method_version,
            experimental_data_identifier=header.experimental_data_identifier,
            container_type=DEFAULT_CONTAINER_TYPE,
            plate_well_count=header.plate_well_count,
            measurements=[
                MapperMeasurement(
                    identifier=measurement.identifier,
                    measurement_time=header.measurement_time,
                    sample_identifier=measurement.sample_identifier,
                    location_identifier=measurement.location_identifier,
                    sample_volume_setting=header.sample_volume_setting,
                    dilution_factor_setting=measurement.dilution_factor_setting,
                    minimum_assay_bead_count_setting=header.minimum_assay_bead_count_setting,
                    detector_gain_setting=header.detector_gain_setting,
                    assay_bead_count=measurement.assay_bead_count,
                    analytes=measurement.analytes,
                    errors=measurement.errors if measurement.errors else None,
                    measurement_custom_info=measurement.measurement_custom_info,
                    sample_custom_info=measurement.sample_custom_info,
                    device_control_custom_info=measurement.device_control_custom_info,
                )
            ],
            custom_info={
                "BatchStopTime": header.custom_info.get("BatchStopTime", None),
                "ProtocolDescription": header.custom_info.get(
                    "ProtocolDescription", None
                ),
            },
        )
        for measurement in measurements
    ]
    # Remove the header custom info that was used to create the measurement groups.
    header.custom_info.pop("BatchStopTime", None)
    header.custom_info.pop("ProtocolDescription", None)

    calculated_documents = list(
        itertools.chain(*[measurement.calculated_data for measurement in measurements])
    )
    return measurement_groups, calculated_documents
